"""End-to-end tests for agentic (agent identity) scenarios.

These tests verify the full agent identity flow using MSAL Python APIs:
1. Assertion callback context propagation (fmi_path flows to callback)
2. Agent app-only token acquisition using FMI-sourced client assertion (Leg 2)
3. Full 3-leg flow: FMI → assertion → user_fic → user-scoped token
4. Cache isolation between app-only and user-scoped tokens

Test configuration uses the same lab infrastructure as test_fmi_e2e.py.
Requires LAB_APP_CLIENT_CERT_PFX_PATH environment variable.
"""

import logging
import os
import sys
import unittest

import msal
from tests.http_client import MinimalHttpClient
from tests.lab_config import get_client_certificate
from tests.test_e2e import LabBasedTestCase

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG if "-v" in sys.argv else logging.INFO)

# =============================================================================
# Test configuration — shared lab app registrations for agentic scenarios
# =============================================================================
_TENANT_ID = "10c419d4-4a50-45b2-aa4e-919fb84df24f"
_BLUEPRINT_CLIENT_ID = "aab5089d-e764-47e3-9f28-cc11c2513821"
_AGENT_APP_ID = "ab18ca07-d139-4840-8b3b-4be9610c6ed5"
_RMA_CLIENT_ID = "3bf56293-fbb5-42bd-a407-248ba7431a8c"
_USER_UPN = "agentuser1@id4slab1.onmicrosoft.com"
_TOKEN_EXCHANGE_SCOPE = "api://AzureADTokenExchange/.default"
_FMI_EXCHANGE_SCOPE = "api://AzureFMITokenExchange/.default"
_GRAPH_SCOPE = "https://graph.microsoft.com/.default"
_FMI_PATH = "SomeFmiPath/FmiCredentialPath"
_AUTHORITY = "https://login.microsoftonline.com/" + _TENANT_ID


# =============================================================================
# Helpers
# =============================================================================

def _acquire_fmi_credential_for_agent(agent_app_id):
    """Leg 1: Blueprint app acquires FMI credential (T1) for the given agent.

    Uses certificate authentication with SNI (sendX5C) and fmi_path set to
    the agent app ID.
    """
    blueprint_app = msal.ConfidentialClientApplication(
        _BLUEPRINT_CLIENT_ID,
        client_credential=get_client_certificate(),
        authority=_AUTHORITY,
        http_client=MinimalHttpClient(),
    )
    result = blueprint_app.acquire_token_for_client(
        [_TOKEN_EXCHANGE_SCOPE], fmi_path=agent_app_id)
    if "access_token" not in result:
        raise RuntimeError(
            "Leg 1 failed — could not acquire FMI credential: {}: {}".format(
                result.get("error"), result.get("error_description")))
    return result["access_token"]


def _acquire_fmi_credential_from_rma():
    """Acquire an FMI credential from the RMA app using certificate credentials.

    Uses the same RMA pattern as test_fmi_e2e._get_fmi_credential_from_rma().
    Used for assertion callback context tests where the callback just needs to
    return a valid FMI token (not specifically for an agent app).
    """
    rma_app = msal.ConfidentialClientApplication(
        _RMA_CLIENT_ID,
        client_credential=get_client_certificate(),
        authority=_AUTHORITY,
        http_client=MinimalHttpClient(),
    )
    result = rma_app.acquire_token_for_client(
        [_FMI_EXCHANGE_SCOPE], fmi_path=_FMI_PATH)
    if "access_token" not in result:
        raise RuntimeError(
            "RMA FMI credential acquisition failed: {}: {}".format(
                result.get("error"), result.get("error_description")))
    return result["access_token"]


def _acquire_instance_token_for_agent():
    """Leg 1 + Leg 2: Acquire an instance token (T2) for the agent app.

    1. Blueprint → T1 (FMI credential via fmi_path)
    2. Agent uses T1 as client_assertion → T2 (instance token)

    T2 is used as user_federated_identity_credential in Leg 3 (user_fic).
    """
    t1 = _acquire_fmi_credential_for_agent(_AGENT_APP_ID)

    agent_app = msal.ConfidentialClientApplication(
        _AGENT_APP_ID,
        client_credential={"client_assertion": t1},
        authority=_AUTHORITY,
        http_client=MinimalHttpClient(),
    )
    result = agent_app.acquire_token_for_client([_TOKEN_EXCHANGE_SCOPE])
    if "access_token" not in result:
        raise RuntimeError(
            "Leg 2 failed — could not acquire instance token: {}: {}".format(
                result.get("error"), result.get("error_description")))
    return result["access_token"]


# =============================================================================
# Tests
# =============================================================================

class TestAssertionCallbackContext(LabBasedTestCase):
    """Verify assertion callback receives correct context when fmi_path is set."""

    def test_assertion_callback_receives_fmi_path(self):
        captured_context = {}

        def assertion_callback(context):
            captured_context.update(context)
            return _acquire_fmi_credential_from_rma()

        app = msal.ConfidentialClientApplication(
            "urn:microsoft:identity:fmi",
            client_credential={"client_assertion": assertion_callback},
            authority=_AUTHORITY,
            http_client=MinimalHttpClient(),
        )

        result = app.acquire_token_for_client(
            [_FMI_EXCHANGE_SCOPE], fmi_path=_AGENT_APP_ID)
        self.assertIn("access_token", result,
            "acquire_token_for_client failed: {}: {}".format(
                result.get("error"), result.get("error_description")))

        # Verify context was passed to callback
        self.assertEqual(_AGENT_APP_ID, captured_context.get("fmi_path"),
            "fmi_path should flow to assertion callback context")
        self.assertEqual("urn:microsoft:identity:fmi", captured_context.get("client_id"),
            "client_id should be in assertion callback context")
        self.assertTrue(captured_context.get("token_endpoint"),
            "token_endpoint should be in assertion callback context")


class TestAgentAppToken(LabBasedTestCase):
    """Agent acquires app-only token for Graph using FMI-sourced assertion.

    Flow: Blueprint → T1 (assertion callback) → Agent CCA → app token

    Disabled in CI: The blueprint app (aab5089d) requires SNI authentication,
    but the CI pipeline's PFX-based cert loading does not include intermediate
    certs in the x5c chain, causing AADSTS700027. These tests pass locally
    where the OS cert store can resolve the chain.
    """

    @unittest.skipUnless(
        os.environ.get("MSAL_RUN_LOCAL_ONLY_TESTS"),
        "Requires local cert store for SNI — set MSAL_RUN_LOCAL_ONLY_TESTS=1")
    def test_agent_gets_app_token_for_graph(self):
        def assertion_provider(context):
            return _acquire_fmi_credential_for_agent(_AGENT_APP_ID)

        agent_app = msal.ConfidentialClientApplication(
            _AGENT_APP_ID,
            client_credential={"client_assertion": assertion_provider},
            authority=_AUTHORITY,
            http_client=MinimalHttpClient(),
        )

        result = agent_app.acquire_token_for_client([_GRAPH_SCOPE])
        self.assertIn("access_token", result,
            "Agent app token acquisition failed: {}: {}".format(
                result.get("error"), result.get("error_description")))
        self.assertTrue(result["access_token"],
            "Access token should not be empty")


class TestAgentUserIdentity(LabBasedTestCase):
    """Full 3-leg agent identity flow: FMI → assertion → user_fic → user token.

    Flow:
    1. Blueprint → T1 (FMI credential)
    2. Agent uses T1 → T2 (instance token)
    3. Agent exchanges T2 via user_fic → user-scoped Graph token
    4. Verify token is cached and retrievable via acquire_token_silent

    Disabled in CI: see TestAgentAppToken docstring.
    """

    @unittest.skipUnless(
        os.environ.get("MSAL_RUN_LOCAL_ONLY_TESTS"),
        "Requires local cert store for SNI — set MSAL_RUN_LOCAL_ONLY_TESTS=1")
    def test_agent_user_identity_gets_token_for_graph(self):
        # Get instance token (T2) for user_fic exchange
        t2 = _acquire_instance_token_for_agent()

        # Build agent CCA with assertion callback
        def assertion_provider(context):
            return _acquire_fmi_credential_for_agent(_AGENT_APP_ID)

        agent_app = msal.ConfidentialClientApplication(
            _AGENT_APP_ID,
            client_credential={"client_assertion": assertion_provider},
            authority=_AUTHORITY,
            http_client=MinimalHttpClient(),
        )

        # Exchange T2 for user-scoped token via user_fic grant
        result = agent_app.acquire_token_by_user_federated_identity_credential(
            [_GRAPH_SCOPE], assertion=t2, username=_USER_UPN)
        self.assertIn("access_token", result,
            "user_fic token acquisition failed: {}: {}".format(
                result.get("error"), result.get("error_description")))
        self.assertTrue(result["access_token"],
            "Access token should not be empty")

        # Verify account was created
        accounts = agent_app.get_accounts()
        self.assertTrue(len(accounts) > 0,
            "Account should be created from user_fic response")

        # Verify silent retrieval works (token should be cached)
        account = accounts[0]
        silent_result = agent_app.acquire_token_silent(
            [_GRAPH_SCOPE], account=account)
        self.assertIsNotNone(silent_result,
            "acquire_token_silent should return cached token")
        self.assertIn("access_token", silent_result)
        self.assertEqual(result["access_token"], silent_result["access_token"],
            "Silent call should return the same cached token")


class TestAgentCacheIsolation(LabBasedTestCase):
    """App-only and user-scoped tokens are isolated in cache on the same CCA.

    Disabled in CI: see TestAgentAppToken docstring.
    """

    @unittest.skipUnless(
        os.environ.get("MSAL_RUN_LOCAL_ONLY_TESTS"),
        "Requires local cert store for SNI — set MSAL_RUN_LOCAL_ONLY_TESTS=1")
    def test_app_and_user_tokens_are_isolated(self):
        def assertion_provider(context):
            return _acquire_fmi_credential_for_agent(_AGENT_APP_ID)

        agent_app = msal.ConfidentialClientApplication(
            _AGENT_APP_ID,
            client_credential={"client_assertion": assertion_provider},
            authority=_AUTHORITY,
            http_client=MinimalHttpClient(),
        )

        # Acquire app-only token
        app_result = agent_app.acquire_token_for_client([_GRAPH_SCOPE])
        self.assertIn("access_token", app_result,
            "App token acquisition failed: {}: {}".format(
                app_result.get("error"), app_result.get("error_description")))

        # Acquire user token via user_fic
        t2 = _acquire_instance_token_for_agent()
        user_result = agent_app.acquire_token_by_user_federated_identity_credential(
            [_GRAPH_SCOPE], assertion=t2, username=_USER_UPN)
        self.assertIn("access_token", user_result,
            "User token acquisition failed: {}: {}".format(
                user_result.get("error"), user_result.get("error_description")))

        # Tokens should be different (app-scoped vs user-scoped)
        self.assertNotEqual(app_result["access_token"], user_result["access_token"],
            "App token and user token should be different")

        # Verify both are independently retrievable
        # App token: second call should return from cache
        app_cached = agent_app.acquire_token_for_client([_GRAPH_SCOPE])
        self.assertEqual("cache", app_cached.get("token_source"),
            "App token should be returned from cache on second call")
        self.assertEqual(app_result["access_token"], app_cached["access_token"])

        # User token: silent call should return from cache
        accounts = agent_app.get_accounts()
        self.assertTrue(len(accounts) > 0)
        user_cached = agent_app.acquire_token_silent(
            [_GRAPH_SCOPE], account=accounts[0])
        self.assertIsNotNone(user_cached)
        self.assertEqual(user_result["access_token"], user_cached["access_token"])


if __name__ == "__main__":
    unittest.main()



import hashlib
import json
import os
import sys
import time
import uuid
from typing import List, Optional
import unittest
try:
    from unittest.mock import patch, ANY, mock_open, Mock
except:
    from mock import patch, ANY, mock_open, Mock
import requests

from tests.test_throttled_http_client import (
    MinimalResponse, ThrottledHttpClientBaseTestCase, DummyHttpClient)
from msal import (
    SystemAssignedManagedIdentity, UserAssignedManagedIdentity,
    ManagedIdentityClient,
    ManagedIdentityError,
    ArcPlatformNotSupportedError,
)
from msal.managed_identity import (
    _ThrottledHttpClient,
    _supported_arc_platforms_and_their_prefixes,
    get_managed_identity_source,
    APP_SERVICE,
    AZURE_ARC,
    CLOUD_SHELL,
    MACHINE_LEARNING,
    SERVICE_FABRIC,
    DEFAULT_TO_VM,
)
from msal.token_cache import is_subdict_of

EXPECTED_SKU = "MSAL.Python"  # Hardcoded constant, not imported from product


class ManagedIdentityTestCase(unittest.TestCase):
    def test_helper_class_should_be_interchangable_with_dict_which_could_be_loaded_from_file_or_env_var(self):
        self.assertEqual(
            UserAssignedManagedIdentity(client_id="foo"),
            {"ManagedIdentityIdType": "ClientId", "Id": "foo"})
        self.assertEqual(
            UserAssignedManagedIdentity(resource_id="foo"),
            {"ManagedIdentityIdType": "ResourceId", "Id": "foo"})
        self.assertEqual(
            UserAssignedManagedIdentity(object_id="foo"),
            {"ManagedIdentityIdType": "ObjectId", "Id": "foo"})
        with self.assertRaises(ManagedIdentityError):
            UserAssignedManagedIdentity()
        with self.assertRaises(ManagedIdentityError):
            UserAssignedManagedIdentity(client_id="foo", resource_id="bar")
        self.assertEqual(
            SystemAssignedManagedIdentity(),
            {"ManagedIdentityIdType": "SystemAssigned", "Id": None})


class ThrottledHttpClientTestCase(ThrottledHttpClientBaseTestCase):
    def test_throttled_http_client_should_not_alter_original_http_client(self):
        self.assertNotAlteringOriginalHttpClient(_ThrottledHttpClient)

    def test_throttled_http_client_should_not_cache_successful_http_response(self):
        http_cache = {}
        http_client=DummyHttpClient(
            status_code=200,
            response_text='{"access_token": "AT", "expires_in": "1234", "resource": "R"}',
            )
        app = ManagedIdentityClient(
            SystemAssignedManagedIdentity(), http_client=http_client, http_cache=http_cache)
        result = app.acquire_token_for_client(resource="R")
        self.assertEqual("AT", result["access_token"])
        self.assertEqual({}, http_cache, "Should not cache successful http response")

    def test_throttled_http_client_should_cache_unsuccessful_http_response(self):
        http_cache = {}
        http_client=DummyHttpClient(
            status_code=400,
            response_headers={"Retry-After": "1"},
            response_text='{"error": "invalid_request"}',
            )
        app = ManagedIdentityClient(
            SystemAssignedManagedIdentity(), http_client=http_client, http_cache=http_cache)
        result = app.acquire_token_for_client(resource="R")
        self.assertEqual("invalid_request", result["error"])
        self.assertNotEqual({}, http_cache, "Should cache unsuccessful http response")
        self.assertCleanPickle(http_cache)


class ClientTestCase(unittest.TestCase):
    maxDiff = None

    def _build_app(
        self,
        *,
        client_capabilities: Optional[List[str]] = None,
    ):
        return ManagedIdentityClient(
            {   # Here we test it with the raw dict form, to test that
                # the client has no hard dependency on ManagedIdentity object
                "ManagedIdentityIdType": "SystemAssigned", "Id": None,
            },
            http_client=requests.Session(),
            client_capabilities=client_capabilities,
            )

    def setUp(self):
        self.app = self._build_app()

    def test_error_out_on_invalid_input(self):
        with self.assertRaises(ManagedIdentityError):
            ManagedIdentityClient({"foo": "bar"}, http_client=requests.Session())
        with self.assertRaises(ManagedIdentityError):
            ManagedIdentityClient(
                {"ManagedIdentityIdType": "undefined", "Id": "foo"},
                http_client=requests.Session())

    def assertCacheStatus(self, app):
        cache = app._token_cache._cache
        self.assertEqual(1, len(cache.get("AccessToken", [])), "Should have 1 AT")
        at = list(cache["AccessToken"].values())[0]
        self.assertEqual(
            app._managed_identity.get("Id", "SYSTEM_ASSIGNED_MANAGED_IDENTITY"),
            at["client_id"],
            "Should have expected client_id")
        self.assertEqual("managed_identity", at["realm"], "Should have expected realm")

    def _test_happy_path(
        self, app, mocked_http, expires_in, *, resource="R", claims_challenge=None,
    ):
        """It tests a normal token request that is expected to hit IdP,
        a subsequent same token request that is expected to hit cache,
        and then a request with claims_challenge that shall hit IdP again.
        """
        result = app.acquire_token_for_client(resource=resource)
        mocked_http.assert_called()
        call_count = mocked_http.call_count
        expected_result = {
            "access_token": "AT",
            "token_type": "Bearer",
        }
        self.assertTrue(
            is_subdict_of(expected_result, result),  # We will test refresh_on later
            "Should obtain a token response")
        self.assertTrue(result["token_source"], "identity_provider")
        self.assertEqual(expires_in, result["expires_in"], "Should have expected expires_in")
        if expires_in >= 7200:
            expected_refresh_on = int(time.time() + expires_in / 2)
            self.assertTrue(
                expected_refresh_on - 1 <= result["refresh_on"] <= expected_refresh_on + 1,
                "Should have a refresh_on time around the middle of the token's life")

        result = app.acquire_token_for_client(resource=resource)
        self.assertCacheStatus(app)
        self.assertEqual("cache", result["token_source"], "Should hit cache")
        self.assertEqual(
            call_count, mocked_http.call_count,
            "No new call to the mocked http should be made for a cache hit")
        self.assertTrue(
            is_subdict_of(expected_result, result),  # We will test refresh_on later
            "Should obtain a token response")
        self.assertTrue(
            expires_in - 5 < result["expires_in"] <= expires_in,
            "Should have similar expires_in")
        if expires_in >= 7200:
            self.assertTrue(
                expected_refresh_on - 5 < result["refresh_on"] <= expected_refresh_on,
                "Should have a refresh_on time around the middle of the token's life")

        result = app.acquire_token_for_client(
            resource=resource, claims_challenge=claims_challenge or "placeholder")
        self.assertEqual("identity_provider", result["token_source"], "Should miss cache")


class VmTestCase(ClientTestCase):

    def _test_happy_path(self) -> callable:
        expires_in = 7890  # We test a bigger than 7200 value here
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "AT", "expires_in": "%s", "resource": "R"}' % expires_in,
        )) as mocked_method:
            super(VmTestCase, self)._test_happy_path(self.app, mocked_method, expires_in)
            return mocked_method

    def test_happy_path_of_vm(self):
        mock_get = self._test_happy_path()
        mock_get.assert_called_with(
            # The last call contained claims_challenge
            # but since IMDS doesn't support token_sha256_to_refresh,
            # the request shall remain the same as before
            'http://169.254.169.254/metadata/identity/oauth2/token',
            params={'api-version': '2018-02-01', 'resource': 'R'},
            headers={
                'Metadata': 'true',
                'x-client-SKU': EXPECTED_SKU,
                'x-client-Ver': ANY,
                'x-ms-client-request-id': ANY,
            },
            )
        # Validate correlation ID is a valid UUID
        corr_id = mock_get.call_args.kwargs["headers"]["x-ms-client-request-id"]
        uuid.UUID(corr_id)

    @patch.object(ManagedIdentityClient, "_ManagedIdentityClient__instance", "MixedCaseHostName")
    def test_happy_path_of_theoretical_mixed_case_hostname(self):
        """Historically, we used to get the host name from socket.getfqdn(),
        which could return a mixed-case host name on Windows.
        Although we no longer use getfqdn(), we still keep this test case to ensure we tolerate it.
        """
        self.test_happy_path_of_vm()

    @patch.dict(os.environ, {"AZURE_POD_IDENTITY_AUTHORITY_HOST": "http://localhost:1234//"})
    def test_happy_path_of_pod_identity(self):
        mock_get = self._test_happy_path()
        mock_get.assert_called_with(
            'http://localhost:1234/metadata/identity/oauth2/token',
            params={'api-version': '2018-02-01', 'resource': 'R'},
            headers={
                'Metadata': 'true',
                'x-client-SKU': EXPECTED_SKU,
                'x-client-Ver': ANY,
                'x-ms-client-request-id': ANY,
            },
            )
        # Validate correlation ID is a valid UUID
        corr_id = mock_get.call_args.kwargs["headers"]["x-ms-client-request-id"]
        uuid.UUID(corr_id)

    def test_vm_error_should_be_returned_as_is(self):
        raw_error = '{"raw": "error format is undefined"}'
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=400,
            text=raw_error,
        )) as mocked_method:
            self.assertEqual(
                json.loads(raw_error), self.app.acquire_token_for_client(resource="R"))
            self.assertEqual({}, self.app._token_cache._cache)

    def test_vm_resource_id_parameter_should_be_msi_res_id(self):
        app = ManagedIdentityClient(
            {"ManagedIdentityIdType": "ResourceId", "Id": "1234"},
            http_client=requests.Session(),
            )
        with patch.object(app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "AT", "expires_in": 3600, "resource": "R"}',
        )) as mocked_method:
            app.acquire_token_for_client(resource="R")
            mocked_method.assert_called_with(
                'http://169.254.169.254/metadata/identity/oauth2/token',
                params={'api-version': '2018-02-01', 'resource': 'R', 'msi_res_id': '1234'},
                headers={
                    'Metadata': 'true',
                    'x-client-SKU': EXPECTED_SKU,
                    'x-client-Ver': ANY,
                    'x-ms-client-request-id': ANY,
                },
                )
            # Validate correlation ID is a valid UUID
            corr_id = mocked_method.call_args.kwargs["headers"]["x-ms-client-request-id"]
            uuid.UUID(corr_id)


@patch.dict(os.environ, {"IDENTITY_ENDPOINT": "http://localhost", "IDENTITY_HEADER": "foo"})
class AppServiceTestCase(ClientTestCase):

    def test_happy_path(self):
        expires_in = 1234
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "AT", "expires_on": "%s", "resource": "R"}' % (
                int(time.time()) + expires_in),
        )) as mocked_method:
            self._test_happy_path(self.app, mocked_method, expires_in)

    def test_app_service_error_should_be_normalized(self):
        raw_error = '{"statusCode": 500, "message": "error content is undefined"}'
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=500,
            text=raw_error,
        )) as mocked_method:
            self.assertEqual({
                "error": "invalid_scope",
                "error_description": "500, error content is undefined",
            }, self.app.acquire_token_for_client(resource="R"))
            self.assertEqual({}, self.app._token_cache._cache)

    def test_app_service_resource_id_parameter_should_be_mi_res_id(self):
        app = ManagedIdentityClient(
            {"ManagedIdentityIdType": "ResourceId", "Id": "1234"},
            http_client=requests.Session(),
            )
        with patch.object(app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "AT", "expires_on": 12345, "resource": "R"}',
        )) as mocked_method:
            app.acquire_token_for_client(resource="R")
            mocked_method.assert_called_with(
                'http://localhost',
                params={'api-version': '2019-08-01', 'resource': 'R', 'mi_res_id': '1234'},
                headers={'X-IDENTITY-HEADER': 'foo', 'Metadata': 'true'},
                )


@patch.dict(os.environ, {"MSI_ENDPOINT": "http://localhost", "MSI_SECRET": "foo"})
class MachineLearningTestCase(ClientTestCase):

    def test_happy_path(self):
        expires_in = 1234
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "AT", "expires_on": "%s", "resource": "R"}' % (
                int(time.time()) + expires_in),
        )) as mocked_method:
            self._test_happy_path(self.app, mocked_method, expires_in)

    def test_machine_learning_error_should_be_normalized(self):
        raw_error = '{"error": "placeholder", "message": "placeholder"}'
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=500,
            text=raw_error,
        )) as mocked_method:
            self.assertEqual({
                "error": "invalid_scope",
                "error_description": "{'error': 'placeholder', 'message': 'placeholder'}",
            }, self.app.acquire_token_for_client(resource="R"))
            self.assertEqual({}, self.app._token_cache._cache)


@patch.dict(os.environ, {
    "IDENTITY_ENDPOINT": "http://localhost",
    "IDENTITY_HEADER": "foo",
    "IDENTITY_SERVER_THUMBPRINT": "bar",
})
class ServiceFabricTestCase(ClientTestCase):
    access_token = "AT"
    access_token_sha256 = hashlib.sha256(access_token.encode()).hexdigest()

    def _test_happy_path(self, app, *, claims_challenge=None) -> callable:
        expires_in = 1234
        with patch.object(app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "%s", "expires_on": %s, "resource": "R", "token_type": "Bearer"}' % (
                self.access_token, int(time.time()) + expires_in),
        )) as mocked_method:
            super(ServiceFabricTestCase, self)._test_happy_path(
                app, mocked_method, expires_in, claims_challenge=claims_challenge)
            return mocked_method

    def test_happy_path_with_client_capabilities_should_relay_capabilities(self):
        self._test_happy_path(self._build_app(client_capabilities=["foo", "bar"])).assert_called_with(
            'http://localhost',
            params={
                'api-version': '2019-07-01-preview',
                'resource': 'R',
                'token_sha256_to_refresh': self.access_token_sha256,
                "xms_cc": "foo,bar",
            },
            headers={'Secret': 'foo'},
        )

    def test_happy_path_with_claim_challenge_should_send_sha256_to_provider(self):
        self._test_happy_path(
            self._build_app(client_capabilities=[]),  # Test empty client_capabilities
            claims_challenge='{"access_token": {"nbf": {"essential": true, "value": "1563308371"}}}',
        ).assert_called_with(
            'http://localhost',
            params={
                'api-version': '2019-07-01-preview',
                'resource': 'R',
                'token_sha256_to_refresh': self.access_token_sha256,
                # There is no xms_cc in this case
            },
            headers={'Secret': 'foo'},
        )

    def test_unified_api_service_should_ignore_unnecessary_client_id(self):
        self._test_happy_path(ManagedIdentityClient(
            {"ManagedIdentityIdType": "ClientId", "Id": "foo"},
            http_client=requests.Session(),
            ))

    def test_sf_error_should_be_normalized(self):
        raw_error = '''
{"error": {
    "correlationId": "foo",
    "code": "SecretHeaderNotFound",
    "message": "Secret is not found in the request headers."
}}'''  # https://learn.microsoft.com/en-us/azure/service-fabric/how-to-managed-identity-service-fabric-app-code#error-handling
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=404,
            text=raw_error,
        )) as mocked_method:
            self.assertEqual({
                "error": "unauthorized_client",
                "error_description": raw_error,
            }, self.app.acquire_token_for_client(resource="R"))
            self.assertEqual({}, self.app._token_cache._cache)


@patch.dict(os.environ, {
    "IDENTITY_ENDPOINT": "http://localhost/token",
    "IMDS_ENDPOINT": "http://localhost",
})
@patch(
    "builtins.open" if sys.version_info.major >= 3 else "__builtin__.open",
    new=mock_open(read_data="secret"),  # `new` requires no extra argument on the decorated function.
        #  https://docs.python.org/3/library/unittest.mock.html#unittest.mock.patch
)
@patch("os.stat", return_value=Mock(st_size=4096))
class ArcTestCase(ClientTestCase):
    challenge = MinimalResponse(status_code=401, text="", headers={
        "WWW-Authenticate": "Basic realm=/tmp/foo",
        })

    def test_error_out_on_invalid_input(self, mocked_stat):
        return super(ArcTestCase, self).test_error_out_on_invalid_input()

    def test_happy_path(self, mocked_stat):
        expires_in = 1234
        with patch.object(self.app._http_client, "get", side_effect=[
            self.challenge,
            MinimalResponse(
                status_code=200,
                text='{"access_token": "AT", "expires_in": "%s", "resource": "R"}' % expires_in,
                ),
            ] * 2,  # Duplicate a pair of mocks for _test_happy_path()'s CAE check
        ) as mocked_method:
            try:
                self._test_happy_path(self.app, mocked_method, expires_in)
                mocked_stat.assert_called_with(os.path.join(
                    _supported_arc_platforms_and_their_prefixes[sys.platform],
                    "foo.key"))
            except ArcPlatformNotSupportedError:
                if sys.platform in _supported_arc_platforms_and_their_prefixes:
                    self.fail("Should not raise ArcPlatformNotSupportedError")

    def test_arc_error_should_be_normalized(self, mocked_stat):
        with patch.object(self.app._http_client, "get", side_effect=[
            self.challenge,
            MinimalResponse(status_code=400, text="undefined"),
        ]) as mocked_method:
            try:
                self.assertEqual({
                    "error": "invalid_request",
                    "error_description": "undefined",
                }, self.app.acquire_token_for_client(resource="R"))
                self.assertEqual({}, self.app._token_cache._cache)
            except ArcPlatformNotSupportedError:
                if sys.platform in _supported_arc_platforms_and_their_prefixes:
                    self.fail("Should not raise ArcPlatformNotSupportedError")


class GetManagedIdentitySourceTestCase(unittest.TestCase):

    @patch.dict(os.environ, {
        "IDENTITY_ENDPOINT": "http://localhost",
        "IDENTITY_HEADER": "foo",
        "IDENTITY_SERVER_THUMBPRINT": "bar",
    })
    def test_service_fabric(self):
        self.assertEqual(get_managed_identity_source(), SERVICE_FABRIC)

    @patch.dict(os.environ, {
        "IDENTITY_ENDPOINT": "http://localhost",
        "IDENTITY_HEADER": "foo",
    })
    def test_app_service(self):
        self.assertEqual(get_managed_identity_source(), APP_SERVICE)

    @patch.dict(os.environ, {
        "MSI_ENDPOINT": "http://localhost",
        "MSI_SECRET": "foo",
    })
    def test_machine_learning(self):
        self.assertEqual(get_managed_identity_source(), MACHINE_LEARNING)

    @patch.dict(os.environ, {
        "IDENTITY_ENDPOINT": "http://localhost",
        "IMDS_ENDPOINT": "http://localhost",
    })
    def test_arc_by_env_var(self):
        self.assertEqual(get_managed_identity_source(), AZURE_ARC)

    @patch("msal.managed_identity.os.path.exists", return_value=True)
    @patch("msal.managed_identity.sys.platform", new="linux")
    def test_arc_by_file_existence_on_linux(self, mocked_exists):
        self.assertEqual(get_managed_identity_source(), AZURE_ARC)
        mocked_exists.assert_called_with("/opt/azcmagent/bin/himds")

    @patch("msal.managed_identity.os.path.exists", return_value=True)
    @patch("msal.managed_identity.sys.platform", new="win32")
    @patch.dict(os.environ, {"ProgramFiles": r"C:\Program Files"})
    def test_arc_by_file_existence_on_windows(self, mocked_exists):
        self.assertEqual(get_managed_identity_source(), AZURE_ARC)
        mocked_exists.assert_called_with(
            r"C:\Program Files\AzureConnectedMachineAgent\himds.exe")

    @patch.dict(os.environ, {
        "AZUREPS_HOST_ENVIRONMENT": "cloud-shell-foo",
    })
    def test_cloud_shell(self):
        self.assertEqual(get_managed_identity_source(), CLOUD_SHELL)

    def test_default_to_vm(self):
        self.assertEqual(get_managed_identity_source(), DEFAULT_TO_VM)




"""If the following ENV VAR were available, many end-to-end test cases would run.
LAB_APP_CLIENT_CERT_PFX_PATH=...
"""
try:
    from dotenv import load_dotenv  # Use this only in local dev machine
    load_dotenv()  # take environment variables from .env.
except:
    pass
import base64
import logging
import os
import json
import time
import unittest
from urllib.parse import urlparse, parse_qs
import sys
try:
    from unittest.mock import patch, ANY
except:
    from mock import patch, ANY

import requests

import msal
from tests.http_client import MinimalHttpClient, MinimalResponse
from msal.oauth2cli import AuthCodeReceiver
from msal.oauth2cli.oidc import decode_part
from tests.broker_util import is_pymsalruntime_installed
from tests.lab_config import (
    get_user_config, get_app_config, get_user_password, get_secret,
    UserSecrets, AppSecrets, LAB_APP_CLIENT_ID, clean_env,
)


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG if "-v" in sys.argv else logging.INFO)

try:
    from dotenv import load_dotenv  # Use this only in local dev machine
    load_dotenv()  # take environment variables from .env.
except ImportError:
    logger.warn("Run pip install -r requirements.txt for optional dependency")

_PYMSALRUNTIME_INSTALLED = is_pymsalruntime_installed()
_AZURE_CLI = "04b07795-8ddb-461a-bbee-02f9e1bf7b46"
# Skip interactive / browser-dependent tests when:
#   - on Travis CI (TRAVIS), or
#   - on Azure DevOps (TF_BUILD) where there is no display/browser on the agent, or
#   - not running in any CI environment at all (not CI).
# Service-principal and ROPC tests are NOT gated on this flag; only tests that
# call acquire_token_interactive() or acquire_token_by_device_flow() are.
_SKIP_UNATTENDED_E2E_TESTS = (
    os.getenv("TRAVIS") or os.getenv("TF_BUILD") or not os.getenv("CI")
)



def _get_app_and_auth_code(
        client_id,
        client_secret=None,
        authority="https://login.microsoftonline.com/common",
        port=44331,
        scopes=["https://graph.microsoft.com/.default"],  # Microsoft Graph
        **kwargs):
    from msal.oauth2cli.authcode import obtain_auth_code
    if client_secret:
        app = msal.ConfidentialClientApplication(
            client_id,
            client_credential=client_secret,
            authority=authority, http_client=MinimalHttpClient())
    else:
        app = msal.PublicClientApplication(
            client_id, authority=authority, http_client=MinimalHttpClient())
    redirect_uri = "http://localhost:%d" % port
    ac = obtain_auth_code(port, auth_uri=app.get_authorization_request_url(
        scopes, redirect_uri=redirect_uri, **kwargs))
    assert ac is not None
    return (app, ac, redirect_uri)

def _render(url, description=None):
    # Render a url in html if description is available, otherwise return url as-is
    return "<a href='{url}' target=_blank>{description}</a>".format(
        url=url, description=description) if description else url


def _get_hint(html_mode=None, username=None, lab_name=None, username_uri=None):
    return "Sign in with {user} whose password is available from {lab}".format(
        user=("<b>{}</b>".format(username) if html_mode else username)
            if username
            else "the upn from {}".format(_render(
                username_uri, description="here" if html_mode else None)),
        lab=_render(
            "https://aka.ms/GetLabSecret?Secret=" + (lab_name or "msidlabXYZ"),
            description="this password api" if html_mode else None,
            ),
        )


@unittest.skipIf(os.getenv("TRAVIS_TAG"), "Skip e2e tests during tagged release")
class E2eTestCase(unittest.TestCase):

    def assertLoosely(self, response, assertion=None,
            skippable_errors=("invalid_grant", "interaction_required")):
        if response.get("error") in skippable_errors:
            logger.debug("Response = %s", response)
            # Some of these errors are configuration issues, not library issues
            raise unittest.SkipTest(response.get("error_description"))
        else:
            if assertion is None:
                assertion = lambda: self.assertIn(
                    "access_token", response,
                    "{error}: {error_description}".format(
                        # Do explicit response.get(...) rather than **response
                        error=response.get("error"),
                        error_description=response.get("error_description")))
            assertion()

    def assertCacheWorksForUser(
            self, result_from_wire, scope, username=None, data=None, auth_scheme=None):
        logger.debug(
            "%s: cache = %s, id_token_claims = %s",
            self.id(),
            json.dumps(self.app.token_cache._cache, indent=4),
            json.dumps(result_from_wire.get("id_token_claims"), indent=4),
            )
        # You can filter by predefined username, or let end user to choose one
        accounts = self.app.get_accounts(username=username)
        self.assertNotEqual(0, len(accounts))
        account = accounts[0]
        if ("scope" not in result_from_wire  # This is the usual case
                or  # Authority server could return different set of scopes
                set(scope) <= set(result_from_wire["scope"].split(" "))
                ):
            # Going to test acquire_token_silent(...) to locate an AT from cache
            silent_result = self.app.acquire_token_silent(
                scope, account=account, data=data or {}, auth_scheme=auth_scheme)
            self.assertIsNotNone(silent_result)
            self.assertIsNone(
                silent_result.get("refresh_token"), "acquire_token_silent() should return no RT")
            if auth_scheme:
                self.assertNotEqual(
                    self.app._TOKEN_SOURCE_CACHE, silent_result[self.app._TOKEN_SOURCE])
            else:
                self.assertEqual(
                    self.app._TOKEN_SOURCE_CACHE, silent_result[self.app._TOKEN_SOURCE])

        if "refresh_token" in result_from_wire:
            assert auth_scheme is None
            # Going to test acquire_token_silent(...) to obtain an AT by a RT from cache
            self.app.token_cache._cache["AccessToken"] = {}  # A hacky way to clear ATs
            silent_result = self.app.acquire_token_silent(
                scope, account=account, data=data or {})
            self.assertIsNotNone(silent_result,
                "We should get a result from acquire_token_silent(...) call")
            self.assertEqual(
                # We used to assert it this way:
                #   result_from_wire['access_token'] != silent_result['access_token']
                # but ROPC in B2C tends to return the same AT we obtained seconds ago.
                # Now looking back, "refresh_token grant would return a brand new AT"
                # was just an empirical observation but never a commitment in specs,
                # so we adjust our way to assert here.
                self.app._TOKEN_SOURCE_IDP, silent_result[self.app._TOKEN_SOURCE])

    def assertCacheWorksForApp(self, result_from_wire, scope):
        logger.debug(
            "%s: cache = %s, id_token_claims = %s",
            self.id(),
            json.dumps(self.app.token_cache._cache, indent=4),
            json.dumps(result_from_wire.get("id_token_claims"), indent=4),
            )
        self.assertIsNone(
            self.app.acquire_token_silent(scope, account=None),
            "acquire_token_silent(..., account=None) shall always return None")
        # Going to test acquire_token_for_client(...) to locate an AT from cache
        silent_result = self.app.acquire_token_for_client(scope)
        self.assertIsNotNone(silent_result)
        self.assertEqual(self.app._TOKEN_SOURCE_CACHE, silent_result[self.app._TOKEN_SOURCE])

    @classmethod
    def _build_app(cls,
            client_id,
            client_credential=None,
            authority="https://login.microsoftonline.com/common",
            oidc_authority=None,
            scopes=["https://graph.microsoft.com/.default"],  # Microsoft Graph
            http_client=None,
            azure_region=None,
            **kwargs):
        if client_credential:
            return msal.ConfidentialClientApplication(
                client_id,
                client_credential=client_credential,
                authority=authority,
                oidc_authority=oidc_authority,
                azure_region=azure_region,
                http_client=http_client or MinimalHttpClient(),
            )
        else:
            # Reuse same test cases, by running them with and without PyMsalRuntime installed
            return msal.PublicClientApplication(
                client_id,
                authority=authority,
                oidc_authority=oidc_authority,
                http_client=http_client or MinimalHttpClient(),
                enable_broker_on_windows=_PYMSALRUNTIME_INSTALLED,
                enable_broker_on_mac=_PYMSALRUNTIME_INSTALLED,
                enable_broker_on_linux=_PYMSALRUNTIME_INSTALLED,
                enable_broker_on_wsl=_PYMSALRUNTIME_INSTALLED,
                )

    def _test_username_password(self,
            authority=None, client_id=None, username=None, password=None, scope=None,
            oidc_authority=None,
            client_secret=None,  # Since MSAL 1.11, confidential client has ROPC too
            azure_region=None,
            http_client=None,
            auth_scheme=None,
            **ignored):
        assert client_id and username and password and scope and (
            authority or oidc_authority)
        self.app = self._build_app(
            client_id, authority=authority, oidc_authority=oidc_authority,
            http_client=http_client,
            azure_region=azure_region,  # Regional endpoint does not support ROPC.
                # Here we just use it to test a regional app won't break ROPC.
            client_credential=client_secret)
        self.assertEqual(
            self.app.get_accounts(username=username), [], "Cache starts empty")
        result = self.app.acquire_token_by_username_password(
            username, password, scopes=scope, auth_scheme=auth_scheme)
        self.assertLoosely(result)
        self.assertCacheWorksForUser(
            result, scope,
            username=username,  # Our implementation works even when "profile" scope was not requested, or when profile claims is unavailable in B2C
            auth_scheme=auth_scheme,
            )
        return result

    @unittest.skipIf(
        _SKIP_UNATTENDED_E2E_TESTS,
        "Although it is doable, we still choose to skip device flow to save time")
    def _test_device_flow(
        self,
        *,
        client_id=None, authority=None, oidc_authority=None, scope=None,
        **ignored
    ):
        assert client_id and scope and (authority or oidc_authority)
        self.app = self._build_app(
            client_id, authority=authority, oidc_authority=oidc_authority)
        flow = self.app.initiate_device_flow(scopes=scope)
        assert "user_code" in flow, "DF does not seem to be provisioned: %s".format(
            json.dumps(flow, indent=4))
        logger.info(flow["message"])

        duration = 60
        logger.info("We will wait up to %d seconds for you to sign in" % duration)
        flow["expires_at"] = min(  # Shorten the time for quick test
            flow["expires_at"], time.time() + duration)
        result = self.app.acquire_token_by_device_flow(flow)
        self.assertLoosely(  # It will skip this test if there is no user interaction
            result,
            assertion=lambda: self.assertIn('access_token', result),
            skippable_errors=self.app.client.DEVICE_FLOW_RETRIABLE_ERRORS)
        if "access_token" not in result:
            self.skipTest("End user did not complete Device Flow in time")
        self.assertCacheWorksForUser(result, scope, username=None)
        result["access_token"] = result["refresh_token"] = "************"
        logger.info(
            "%s obtained tokens: %s", self.id(), json.dumps(result, indent=4))

    @unittest.skipIf(_SKIP_UNATTENDED_E2E_TESTS, "Browser automation is not yet implemented")
    def _test_acquire_token_interactive(
            self, *, client_id=None, authority=None, scope=None, port=None,
            oidc_authority=None,
            username=None, lab_name=None,
            username_uri="",  # Unnecessary if you provided username and lab_name
            data=None,  # Needed by ssh-cert feature
            prompt=None,
            enable_msa_passthrough=None,
            auth_scheme=None,
            **ignored):
        assert client_id and scope and (authority or oidc_authority)
        self.app = self._build_app(
            client_id, authority=authority, oidc_authority=oidc_authority)
        logger.info(_get_hint(  # Useful when testing broker which shows no welcome_template
            username=username, lab_name=lab_name, username_uri=username_uri))
        result = self.app.acquire_token_interactive(
            scope,
            login_hint=username,
            prompt=prompt,
            timeout=120,
            port=port,
            parent_window_handle=self.app.CONSOLE_WINDOW_HANDLE,  # This test app is a console app
            enable_msa_passthrough=enable_msa_passthrough,  # Needed when testing MSA-PT app
            welcome_template=  # This is an undocumented feature for testing
                """<html><body><h1>{id}</h1><ol>
    <li>{hint}</li>
    <li><a href="$auth_uri">Sign In</a> or <a href="$abort_uri">Abort</a></li>
    </ol></body></html>""".format(id=self.id(), hint=_get_hint(
                html_mode=True,
                username=username, lab_name=lab_name, username_uri=username_uri)),
            auth_scheme=auth_scheme,
            data=data or {},
            )
        self.assertIn(
            "access_token", result,
            "{error}: {error_description}".format(
                # Note: No interpolation here, cause error won't always present
                error=result.get("error"),
                error_description=result.get("error_description")))
        if username and result.get("id_token_claims", {}).get("preferred_username"):
            self.assertEqual(
                username, result["id_token_claims"]["preferred_username"],
                "You are expected to sign in as account {}, but tokens returned is for {}".format(
                    username, result["id_token_claims"]["preferred_username"]))
        self.assertCacheWorksForUser(
            result, scope, username=None, data=data or {}, auth_scheme=auth_scheme)
        return result  # For further testing


@unittest.skipUnless(
    msal.application._is_running_in_cloud_shell(),
    "Manually run this test case from inside Cloud Shell")
class CloudShellTestCase(E2eTestCase):
    scope_that_requires_no_managed_device = "https://management.core.windows.net/"  # Scopes came from https://msazure.visualstudio.com/One/_git/compute-CloudShell?path=/src/images/agent/env/envconfig.PROD.json&version=GBmaster&_a=contents

    def setUpClass(cls):
        # Doing it here instead of as a class member,
        # otherwise its overhead incurs even when running other cases
        cls.app = msal.PublicClientApplication("client_id")

    def test_access_token_should_be_obtained_for_a_supported_scope(self):
        result = self.app.acquire_token_interactive(
            [self.scope_that_requires_no_managed_device], prompt="none")
        self.assertEqual(
            "Bearer", result.get("token_type"), "Unexpected result: %s" % result)
        self.assertIsNotNone(result.get("access_token"))


class PublicCloudScenariosTestCase(E2eTestCase):
    # Historically this class was driven by tests/config.json for semi-automated runs.
    # It now uses lab config + env vars so it can run automatically on any CI
    # (including Azure DevOps) as long as LAB_APP_CLIENT_CERT_PFX_PATH is set.

    @classmethod
    def setUpClass(cls):
        if not clean_env("LAB_APP_CLIENT_CERT_PFX_PATH"):
            raise unittest.SkipTest(
                "LAB_APP_CLIENT_CERT_PFX_PATH not set; skipping PublicCloud e2e tests")
        pca_app = get_app_config(AppSecrets.PCA_CLIENT)
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        cls.config = {
            "client_id": pca_app.app_id,
            "authority": user.authority,
            "username": user.upn,
            "password": get_user_password(user),
            "scope": ["https://graph.microsoft.com/.default"],
            "listen_port": 44331,
        }

    def skipUnlessWithConfig(self, fields):
        for field in fields:
            if field not in self.config:
                self.skipTest('Skipping due to lack of configuration "%s"' % field)

    def test_username_password(self):
        self.skipUnlessWithConfig(["client_id", "username", "password", "scope"])
        self._test_username_password(**self.config)

    def _get_app_and_auth_code(self, scopes=None, **kwargs):
        return _get_app_and_auth_code(
            self.config["client_id"],
            client_secret=self.config.get("client_secret"),
            authority=self.config.get("authority"),
            port=self.config.get("listen_port", 44331),
            scopes=scopes or self.config["scope"],
            **kwargs)

    def _test_auth_code(self, auth_kwargs, token_kwargs):
        self.skipUnlessWithConfig(["client_id", "scope"])
        (self.app, ac, redirect_uri) = self._get_app_and_auth_code(**auth_kwargs)
        result = self.app.acquire_token_by_authorization_code(
            ac, self.config["scope"], redirect_uri=redirect_uri, **token_kwargs)
        logger.debug("%s.cache = %s",
            self.id(), json.dumps(self.app.token_cache._cache, indent=4))
        self.assertIn(
            "access_token", result,
            "{error}: {error_description}".format(
                # Note: No interpolation here, cause error won't always present
                error=result.get("error"),
                error_description=result.get("error_description")))
        self.assertCacheWorksForUser(result, self.config["scope"], username=None)

    def manual_test_auth_code(self):
        if _SKIP_UNATTENDED_E2E_TESTS:
            self.skipTest("Browser automation is not yet implemented")
        self._test_auth_code({}, {})

    def manual_test_auth_code_with_matching_nonce(self):
        if _SKIP_UNATTENDED_E2E_TESTS:
            self.skipTest("Browser automation is not yet implemented")
        self._test_auth_code({"nonce": "foo"}, {"nonce": "foo"})

    def manual_test_auth_code_with_mismatching_nonce(self):
        if _SKIP_UNATTENDED_E2E_TESTS:
            self.skipTest("Browser automation is not yet implemented")
        self.skipUnlessWithConfig(["client_id", "scope"])
        (self.app, ac, redirect_uri) = self._get_app_and_auth_code(nonce="foo")
        with self.assertRaises(ValueError):
            self.app.acquire_token_by_authorization_code(
                ac, self.config["scope"], redirect_uri=redirect_uri, nonce="bar")

    def test_client_secret(self):
        app = get_app_config(AppSecrets.S2S_CLIENT)
        secret_name = app.secret_name or app.client_secret
        if not (app.app_id and secret_name):
            self.skipTest("S2S app configuration is incomplete")
        self.app = msal.ConfidentialClientApplication(
            app.app_id,
            client_credential=get_secret(secret_name, vault="msal_team"),
            authority=app.authority,
            http_client=MinimalHttpClient())
        scope = ["https://graph.microsoft.com/.default"]
        result = self.app.acquire_token_for_client(scope)
        self.assertIn('access_token', result)
        self.assertCacheWorksForApp(result, scope)

    def test_subject_name_issuer_authentication(self):
        from tests.lab_config import get_client_certificate
        if not clean_env("LAB_APP_CLIENT_CERT_PFX_PATH"):
            self.skipTest("LAB_APP_CLIENT_CERT_PFX_PATH not set")

        self.app = msal.ConfidentialClientApplication(
            LAB_APP_CLIENT_ID,
            authority="https://login.microsoftonline.com/microsoft.onmicrosoft.com",
            client_credential=get_client_certificate(),
            http_client=MinimalHttpClient())
        scope = ["https://graph.microsoft.com/.default"]
        result = self.app.acquire_token_for_client(scope)
        self.assertIn('access_token', result)
        self.assertCacheWorksForApp(result, scope)

class DeviceFlowTestCase(E2eTestCase):  # A leaf class so it will be run only once
    @classmethod
    def setUpClass(cls):
        app = get_app_config(AppSecrets.PCA_CLIENT)
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        cls.config = {
            "client_id": app.app_id,
            "authority": user.authority,
            "scope": ["https://graph.microsoft.com/.default"],
        }

    def manual_test_device_flow(self):
        self._test_device_flow(**self.config)


def get_lab_app(
        env_client_cert_path="LAB_APP_CLIENT_CERT_PFX_PATH",
        authority="https://login.microsoftonline.com/"
            "72f988bf-86f1-41af-91ab-2d7cd011db47",  # Microsoft tenant ID
        timeout=None,
        **kwargs):
    """Returns the lab app as an MSAL confidential client.

    Uses the hardcoded lab app client ID (RequestMSIDLAB) and a certificate
    from the LAB_APP_CLIENT_CERT_PFX_PATH env var.
    """
    logger.info(
        "Reading ENV variable %s for lab app defined at "
        "https://docs.msidlab.com/accounts/confidentialclient.html",
        env_client_cert_path)
    cert_path = clean_env(env_client_cert_path)
    if cert_path:
        # id came from https://docs.msidlab.com/accounts/confidentialclient.html
        client_credential = {
            "private_key_pfx_path":
                # Cert came from https://ms.portal.azure.com/#@microsoft.onmicrosoft.com/asset/Microsoft_Azure_KeyVault/Certificate/https://msidlabs.vault.azure.net/certificates/LabAuth
                cert_path,
            "public_certificate": True,  # Opt in for SNI
            }
    else:
        logger.info("ENV variables are not defined. Fall back to MSI.")
        # See also https://microsoft.sharepoint-df.com/teams/MSIDLABSExtended/SitePages/Programmatically-accessing-LAB-API's.aspx
        raise unittest.SkipTest("MSI-based mechanism has not been implemented yet")
    return msal.ConfidentialClientApplication(
            LAB_APP_CLIENT_ID,
            client_credential=client_credential,
            authority=authority,
            http_client=MinimalHttpClient(timeout=timeout),
            **kwargs)

class LabTokenError(RuntimeError):
    pass

def get_session(lab_app, scopes):  # BTW, this infrastructure tests the confidential client flow
    logger.info("Creating session")
    result = lab_app.acquire_token_for_client(scopes)
    if not result.get("access_token"):
        raise LabTokenError(
            "Unable to obtain token for lab. Encountered {}: {}".format(
            result.get("error"), result.get("error_description")
        ))
    session = requests.Session()
    session.headers.update({"Authorization": "Bearer %s" % result["access_token"]})
    session.hooks["response"].append(lambda r, *args, **kwargs: r.raise_for_status())
    return session


class LabBasedTestCase(E2eTestCase):
    _secrets = {}
    adfs2019_scopes = ["placeholder"]  # Need this to satisfy MSAL API surface.
        # Internally, MSAL will also append more scopes like "openid" etc..
        # ADFS 2019 will issue tokens for valid scope only, by default "openid".
        # https://docs.microsoft.com/en-us/windows-server/identity/ad-fs/overview/ad-fs-faq#what-permitted-scopes-are-supported-by-ad-fs

    @classmethod
    def setUpClass(cls):
        # https://docs.msidlab.com/accounts/apiaccess.html#code-snippet
        cls.session = get_session(get_lab_app(), [
                "https://request.msidlab.com/.default",  # A lab change since June 10, 2024
            ])

    @classmethod
    def tearDownClass(cls):
        cls.session.close()

    @unittest.skipIf(_SKIP_UNATTENDED_E2E_TESTS, "Browser automation is not yet implemented")
    def _test_acquire_token_by_auth_code(
            self, client_id=None, authority=None, port=None, scope=None,
            **ignored):
        assert client_id and authority and port and scope
        (self.app, ac, redirect_uri) = _get_app_and_auth_code(
            client_id, authority=authority, port=port, scopes=scope)
        result = self.app.acquire_token_by_authorization_code(
            ac, scope, redirect_uri=redirect_uri)
        logger.debug(
            "%s: cache = %s, id_token_claims = %s",
            self.id(),
            json.dumps(self.app.token_cache._cache, indent=4),
            json.dumps(result.get("id_token_claims"), indent=4),
            )
        self.assertIn(
            "access_token", result,
            "{error}: {error_description}".format(
                # Note: No interpolation here, cause error won't always present
                error=result.get("error"),
                error_description=result.get("error_description")))
        self.assertCacheWorksForUser(result, scope, username=None)

    @unittest.skipIf(_SKIP_UNATTENDED_E2E_TESTS, "Browser automation is not yet implemented")
    def _test_acquire_token_by_auth_code_flow(
            self, client_id=None, authority=None, port=None, scope=None,
            username=None, lab_name=None,
            username_uri="",  # Optional if you provided username and lab_name
            **ignored):
        assert client_id and authority and scope
        self.app = msal.ClientApplication(
            client_id, authority=authority, http_client=MinimalHttpClient())
        with AuthCodeReceiver(port=port) as receiver:
            flow = self.app.initiate_auth_code_flow(
                scope,
                redirect_uri="http://localhost:%d" % receiver.get_port(),
                )
            auth_response = receiver.get_auth_response(
                auth_uri=flow["auth_uri"], state=flow["state"], timeout=60,
                welcome_template="""<html><body><h1>{id}</h1><ol>
    <li>{hint}</li>
    <li><a href="$auth_uri">Sign In</a> or <a href="$abort_uri">Abort</a></li>
    </ol></body></html>""".format(id=self.id(), hint=_get_hint(
                    html_mode=True,
                    username=username, lab_name=lab_name, username_uri=username_uri)),
                )
        if auth_response is None:
            self.skipTest("Timed out. Did not have test settings in hand? Prepare and retry.")
        self.assertIsNotNone(
            auth_response.get("code"), "Error: {}, Detail: {}".format(
                auth_response.get("error"), auth_response))
        result = self.app.acquire_token_by_auth_code_flow(flow, auth_response)
        logger.debug(
            "%s: cache = %s, id_token_claims = %s",
            self.id(),
            json.dumps(self.app.token_cache._cache, indent=4),
            json.dumps(result.get("id_token_claims"), indent=4),
            )
        self.assertIn(
            "access_token", result,
            "{error}: {error_description}".format(
                # Note: No interpolation here, cause error won't always present
                error=result.get("error"),
                error_description=result.get("error_description")))
        if username and result.get("id_token_claims", {}).get("preferred_username"):
            self.assertEqual(
                username, result["id_token_claims"]["preferred_username"],
                "You are expected to sign in as account {}, but tokens returned is for {}".format(
                    username, result["id_token_claims"]["preferred_username"]))
        self.assertCacheWorksForUser(result, scope, username=None)

    def _test_acquire_token_obo(self, config_pca, config_cca,
            azure_region=None,  # Regional endpoint does not really support OBO.
                # Here we just test regional apps won't adversely break OBO
            http_client=None,
            ):
        if "client_secret" not in config_pca:
            # 1.a An app obtains a token representing a user, for our mid-tier service
            result = msal.PublicClientApplication(
                config_pca["client_id"], authority=config_pca["authority"],
                azure_region=azure_region,
                http_client=http_client or MinimalHttpClient(),
                ).acquire_token_by_username_password(
                    config_pca["username"], config_pca["password"],
                    scopes=config_pca["scope"],
                )
        else:  # We repurpose the config_pca to contain client_secret for cca app 1
            # 1.b An app obtains a token representing itself, for our mid-tier service
            result = msal.ConfidentialClientApplication(
                config_pca["client_id"], authority=config_pca["authority"],
                client_credential=config_pca["client_secret"],
                azure_region=azure_region,
                http_client=http_client or MinimalHttpClient(),
                ).acquire_token_for_client(scopes=config_pca["scope"])
        assertion = result.get("access_token")
        self.assertIsNotNone(assertion, "First app failed to get AT. {}".format(
            json.dumps(result, indent=2)))

        # 2. Our mid-tier service uses OBO to obtain a token for downstream service
        cca = msal.ConfidentialClientApplication(
            config_cca["client_id"],
            client_credential=config_cca["client_secret"],
            authority=config_cca["authority"],
            azure_region=azure_region,
            http_client=http_client or MinimalHttpClient(),
            # token_cache= ...,  # Default token cache is all-tokens-store-in-memory.
                # That's fine if OBO app uses short-lived msal instance per session.
                # Otherwise, the OBO app need to implement a one-cache-per-user setup.
            )
        cca_result = cca.acquire_token_on_behalf_of(assertion, config_cca["scope"])
        self.assertIsNotNone(cca_result.get("access_token"), "OBO call failed: {}".format(
            json.dumps(cca_result, indent=2)))

        # 3. Now the OBO app can simply store downstream token(s) in same session.
        #    Alternatively, if you want to persist the downstream AT, and possibly
        #    the RT (if any) for prolonged access even after your own AT expires,
        #    now it is the time to persist current cache state for current user.
        #    Assuming you already did that (which is not shown in this test case),
        #    the following part shows one of the ways to obtain an AT from cache.
        username = cca_result.get("id_token_claims", {}).get("preferred_username")
        accounts = cca.get_accounts(username=username)
        if username is not None:  # It means CCA have requested an IDT w/ "profile" scope
            assert config_cca["username"] == username, "Incorrect test case configuration"
            self.assertEqual(1, len(accounts), "App is supposed to partition token cache per user")
        account = accounts[0]  # Alternatively, cca app could just loop through each account
        result = cca.acquire_token_silent(config_cca["scope"], account)
        self.assertTrue(
            result and result.get("access_token") == cca_result["access_token"],
            "CCA should hit an access token from cache: {}".format(
                json.dumps(cca.token_cache._cache, indent=2)))
        if "refresh_token" in cca_result:
            result = cca.acquire_token_silent(
                config_cca["scope"], account=account, force_refresh=True)
            self.assertTrue(
                result and "access_token" in result,
                "CCA should get an AT silently, but we got this instead: {}".format(result))
            self.assertNotEqual(
                result["access_token"], cca_result["access_token"],
                "CCA should get a new AT")
        else:
            logger.info("AAD did not issue a RT for OBO flow")

    def _test_acquire_token_by_client_secret(
            self, client_id=None, client_secret=None, authority=None, scope=None,
            oidc_authority=None,
            **ignored):
        assert client_id and client_secret and scope and (
            authority or oidc_authority)
        self.app = msal.ConfidentialClientApplication(
            client_id, client_credential=client_secret, authority=authority,
            oidc_authority=oidc_authority,
            http_client=MinimalHttpClient())
        result = self.app.acquire_token_for_client(scope)
        self.assertIsNotNone(result.get("access_token"), "Got %s instead" % result)
        self.assertCacheWorksForApp(result, scope)

    def _test_acquire_token_by_client_cert(
            self, client_id=None, authority=None, scope=None,
            oidc_authority=None,
            **ignored):
        """Test client credentials flow with certificate.
        
        Uses the lab certificate from LAB_APP_CLIENT_CERT_PFX_PATH environment
        variable for authentication.
        """
        from tests.lab_config import get_client_certificate
        
        assert client_id and scope and (authority or oidc_authority)
        client_credential = get_client_certificate()
        self.app = msal.ConfidentialClientApplication(
            client_id, client_credential=client_credential, authority=authority,
            oidc_authority=oidc_authority,
            http_client=MinimalHttpClient())
        result = self.app.acquire_token_for_client(scope)
        self.assertIsNotNone(result.get("access_token"), "Got %s instead" % result)
        self.assertCacheWorksForApp(result, scope)


class PopWithExternalKeyTestCase(LabBasedTestCase):
    """Base class for POP/SSH-cert tests with external key.
    
    Uses the S2S app configuration from Key Vault for service principal tests
    and the public cloud user for user account tests.
    """

    def _test_service_principal(self):
        """Test acquiring POP/SSH-cert token for a service principal.
        
        Uses the S2S app configuration and lab certificate.
        """
        from tests.lab_config import get_app_config, get_client_certificate, AppSecrets
        
        app_config = get_app_config(AppSecrets.S2S_CLIENT)
        client_credential = get_client_certificate()
        
        app = msal.ConfidentialClientApplication(
            app_config.app_id,
            client_credential=client_credential,
            authority=app_config.authority or "https://login.microsoftonline.com/microsoft.onmicrosoft.com",
            http_client=MinimalHttpClient(),
        )
        
        result = app.acquire_token_for_client(self.SCOPE, data=self.DATA1)
        self.assertIsNotNone(result.get("access_token"), "Encountered {}: {}".format(
            result.get("error"), result.get("error_description")))
        self.assertEqual(self.EXPECTED_TOKEN_TYPE, result["token_type"])
        self.assertEqual(result["token_source"], "identity_provider")

        # Test cache hit
        cached_result = app.acquire_token_for_client(self.SCOPE, data=self.DATA1)
        self.assertIsNotNone(
            cached_result.get("access_token"), "Encountered {}: {}".format(
            cached_result.get("error"), cached_result.get("error_description")))
        self.assertEqual(self.EXPECTED_TOKEN_TYPE, cached_result["token_type"])
        self.assertEqual(cached_result["token_source"], "cache")

        # refresh_token grant can fetch an ssh-cert bound to a different key
        refreshed_result = app.acquire_token_for_client(self.SCOPE, data=self.DATA2)
        self.assertIsNotNone(
            refreshed_result.get("access_token"), "Encountered {}: {}".format(
            refreshed_result.get("error"), refreshed_result.get("error_description")))
        self.assertEqual(self.EXPECTED_TOKEN_TYPE, refreshed_result["token_type"])
        self.assertEqual(refreshed_result["token_source"], "identity_provider")

    def _test_user_account(self):
        """Test acquiring POP/SSH-cert token for a user account.
        
        Uses Azure CLI client ID since it's one of the only clients that are
        pre-authorized to use the SSH cert feature.
        """
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        result = self._test_acquire_token_interactive(
            client_id="04b07795-8ddb-461a-bbee-02f9e1bf7b46",  # Azure CLI is one
                # of the only 2 clients that are PreAuthz to use ssh cert feature
            authority="https://login.microsoftonline.com/common",
            scope=self.SCOPE,
            data=self.DATA1,
            username=user.upn,
            lab_name=user.lab_name,
            prompt="none" if msal.application._is_running_in_cloud_shell() else None,
            )   # It already tests reading AT from cache, and using RT to refresh
                # acquire_token_silent() would work because we pass in the same key
        self.assertIsNotNone(result.get("access_token"), "Encountered {}: {}".format(
            result.get("error"), result.get("error_description")))
        self.assertEqual(self.EXPECTED_TOKEN_TYPE, result["token_type"])
        self.assertEqual(result["token_source"], "identity_provider")
        logger.debug("%s.cache = %s",
            self.id(), json.dumps(self.app.token_cache._cache, indent=4))

        # refresh_token grant can hit an ssh-cert bound to the same key
        account = self.app.get_accounts()[0]
        cached_result = self.app.acquire_token_silent(
            self.SCOPE, account=account, data=self.DATA1)
        self.assertIsNotNone(cached_result)
        self.assertEqual(self.EXPECTED_TOKEN_TYPE, cached_result["token_type"])
        ## Actually, the self._test_acquire_token_interactive() already contained
        ## a built-in refresh test, so the token in cache has been refreshed already.
        ## Therefore, the following line won't pass, which is expected.
        #self.assertEqual(result["access_token"], cached_result['access_token'])
        self.assertEqual(cached_result["token_source"], "cache")

        # refresh_token grant can fetch an ssh-cert bound to a different key
        account = self.app.get_accounts()[0]
        refreshed_result = self.app.acquire_token_silent(
            self.SCOPE, account=account, data=self.DATA2)
        self.assertIsNotNone(refreshed_result)
        self.assertEqual(self.EXPECTED_TOKEN_TYPE, refreshed_result["token_type"])
        self.assertNotEqual(result["access_token"], refreshed_result['access_token'])
        self.assertEqual(refreshed_result["token_source"], "identity_provider")


class SshCertTestCase(PopWithExternalKeyTestCase):
    EXPECTED_TOKEN_TYPE = "ssh-cert"
    _JWK1 = """{"kty":"RSA", "n":"2tNr73xwcj6lH7bqRZrFzgSLj7OeLfbn8216uOMDHuaZ6TEUBDN8Uz0ve8jAlKsP9CQFCSVoSNovdE-fs7c15MxEGHjDcNKLWonznximj8pDGZQjVdfK-7mG6P6z-lgVcLuYu5JcWU_PeEqIKg5llOaz-qeQ4LEDS4T1D2qWRGpAra4rJX1-kmrWmX_XIamq30C9EIO0gGuT4rc2hJBWQ-4-FnE1NXmy125wfT3NdotAJGq5lMIfhjfglDbJCwhc8Oe17ORjO3FsB5CLuBRpYmP7Nzn66lRY3Fe11Xz8AEBl3anKFSJcTvlMnFtu3EpD-eiaHfTgRBU7CztGQqVbiQ", "e":"AQAB"}"""
    _JWK2 = """{"kty":"RSA", "n":"72u07mew8rw-ssw3tUs9clKstGO2lvD7ZNxJU7OPNKz5PGYx3gjkhUmtNah4I4FP0DuF1ogb_qSS5eD86w10Wb1ftjWcoY8zjNO9V3ph-Q2tMQWdDW5kLdeU3-EDzc0HQeou9E0udqmfQoPbuXFQcOkdcbh3eeYejs8sWn3TQprXRwGh_TRYi-CAurXXLxQ8rp-pltUVRIr1B63fXmXhMeCAGwCPEFX9FRRs-YHUszUJl9F9-E0nmdOitiAkKfCC9LhwB9_xKtjmHUM9VaEC9jWOcdvXZutwEoW2XPMOg0Ky-s197F9rfpgHle2gBrXsbvVMvS0D-wXg6vsq6BAHzQ", "e":"AQAB"}"""
    DATA1 = {"token_type": "ssh-cert", "key_id": "key1", "req_cnf": _JWK1}
    DATA2 = {"token_type": "ssh-cert", "key_id": "key2", "req_cnf": _JWK2}
    _SCOPE_USER = ["https://pas.windows.net/CheckMyAccess/Linux/user_impersonation"]
    _SCOPE_SP = ["https://pas.windows.net/CheckMyAccess/Linux/.default"]
    SCOPE = _SCOPE_SP  # Historically there was a separation, at 2021 it is unified

    def test_service_principal(self):
        self._test_service_principal()

    def test_user_account(self):
        self._test_user_account()


def _data_for_pop(key):
    raw_req_cnf = json.dumps({"kid": key, "xms_ksl": "sw"})
    return {  # Sampled from Azure CLI's plugin connectedk8s
        'token_type': 'pop',
        'key_id': key,
        "req_cnf": base64.urlsafe_b64encode(raw_req_cnf.encode('utf-8')).decode('utf-8').rstrip('='),
            # Note: Sending raw_req_cnf without base64 encoding would result in an http 500 error
    }  # See also https://github.com/Azure/azure-cli-extensions/blob/main/src/connectedk8s/azext_connectedk8s/_clientproxyutils.py#L86-L92


class AtPopWithExternalKeyTestCase(PopWithExternalKeyTestCase):
    EXPECTED_TOKEN_TYPE = "pop"
    DATA1 = _data_for_pop('AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA-AAAAAAAA')  # Fake key with a certain format and length
    DATA2 = _data_for_pop('BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB-BBBBBBBB')  # Fake key with a certain format and length
    SCOPE = [
        '6256c85f-0aad-4d50-b960-e6e9b21efe35/.default',  # Azure CLI's connectedk8s plugin uses this
            # https://github.com/Azure/azure-cli-extensions/pull/4468/files#diff-a47efa3186c7eb4f1176e07d0b858ead0bf4a58bfd51e448ee3607a5b4ef47f6R116
    ]

    def test_service_principal(self):
        self._test_service_principal()

    def test_user_account(self):
        self._test_user_account()


class WorldWideTestCase(LabBasedTestCase):

    def test_aad_managed_user(self):  # Pure cloud
        """Test username/password flow for a managed AAD user."""
        app = get_app_config(AppSecrets.PCA_CLIENT)
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        password = get_user_password(user)
        self._test_username_password(
            authority=user.authority,
            client_id=app.app_id,
            username=user.upn,
            password=password,
            scope=["https://graph.microsoft.com/.default"],
        )

    def test_adfs2022_fed_user(self):
        """Test username/password flow for a federated user via ADFS 2022."""
        app = get_app_config(AppSecrets.PCA_CLIENT)
        user = get_user_config(UserSecrets.FEDERATED)
        password = get_user_password(user)
        self._test_username_password(
            authority=user.authority,
            client_id=app.app_id,
            username=user.upn,
            password=password,
            scope=["https://graph.microsoft.com/.default"],
        )

    def manual_test_cloud_acquire_token_interactive(self):
        """Test interactive flow for a cloud user."""
        app = get_app_config(AppSecrets.PCA_CLIENT)
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        self._test_acquire_token_interactive(
            authority=user.authority,
            client_id=app.app_id,
            username=user.upn,
            lab_name=user.lab_name,
            scope=["https://graph.microsoft.com/.default"],
        )

    def manual_test_msa_pt_app_signin_via_organizations_authority_without_login_hint(self):
        """There is/was an upstream bug. See test case full docstring for the details.

        When a MSAL-PT flow that account control is launched, user has 2+ AAD accounts in WAM,
        selects an AAD account that is NOT the default AAD account from the OS,
        it will incorrectly get tokens for default AAD account.
        """
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        self._test_acquire_token_interactive(
            authority="https://login.microsoftonline.com/organizations",
            client_id="04b07795-8ddb-461a-bbee-02f9e1bf7b46",  # Azure CLI is an MSA-PT app
            username=user.upn,
            lab_name=user.lab_name,
            scope=["https://graph.microsoft.com/.default"],
            enable_msa_passthrough=True,
            prompt="select_account",  # In MSAL Python, this resets login_hint
        )

    def test_acquire_token_obo(self):
        """Test On-Behalf-Of flow.
        
        Flow:
        1. PCA acquires token for user to access the WebAPI (scope: api://<app_id>/access_as_user)
        2. WebAPI (CCA) uses that token as assertion to get token for downstream service (Graph)
        """
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        password = get_user_password(user)
        web_api_app = get_app_config(AppSecrets.WEB_API_CLIENT)

        # Step 1: PCA gets token for user to access the WebAPI
        # Note: Java test uses "organizations" authority for PCA
        config_pca = {
            "authority": "https://login.microsoftonline.com/organizations",
            "client_id": web_api_app.app_id,
            "username": user.upn,
            "password": password,
            "scope": ["api://%s/access_as_user" % web_api_app.app_id],
        }

        # Step 2: WebAPI (CCA) exchanges the token via OBO for Graph access
        # Note: web_api_app.client_secret contains the Key Vault secret name,
        # which we pass to get_secret() to retrieve the actual secret value.
        config_cca = {
            "authority": user.authority,  # Tenant-specific authority
            "client_id": web_api_app.app_id,
            "client_secret": get_secret(web_api_app.client_secret, vault="msal_team"),
            "scope": ["https://graph.microsoft.com/.default"],
            "username": user.upn,
        }

        self._test_acquire_token_obo(config_pca, config_cca)


    @unittest.skipUnless(
        os.path.exists("tests/sp_obo.pem"),
        "Need a 'tests/sp_obo.pem' private key to run OBO for SP test")
    def test_acquire_token_obo_for_sp(self):
        """Test On-Behalf-Of flow for service principal.
        
        Note: This test uses hardcoded PPE (pre-production environment) values
        as it tests a specific SP-to-SP OBO scenario.
        """
        authority = "https://login.windows-ppe.net/f686d426-8d16-42db-81b7-ab578e110ccd"
        with open("tests/sp_obo.pem") as pem:
            client_secret = {
                "private_key": pem.read(),
                "thumbprint": "378938210C976692D7F523B8C4FFBB645D17CE92",
            }
        midtier_app = {
            "authority": authority,
            "client_id": "c84e9c32-0bc9-4a73-af05-9efe9982a322",
            "client_secret": client_secret,
            "scope": ["23d08a1e-1249-4f7c-b5a5-cb11f29b6923/.default"],
        }
        initial_app = {
            "authority": authority,
            "client_id": "9793041b-9078-4942-b1d2-babdc472cc0c",
            "client_secret": client_secret,
            "scope": [midtier_app["client_id"] + "/.default"],
        }
        self._test_acquire_token_obo(initial_app, midtier_app)

    def test_acquire_token_by_client_secret(self):
        """Test client credentials flow with client secret."""
        app = get_app_config(AppSecrets.S2S_CLIENT)

        client_secret = get_secret(app.secret_name, vault="msal_team")
        self._test_acquire_token_by_client_secret(
            client_id=app.app_id,
            client_secret=client_secret,
            authority=app.authority,
            scope=["https://graph.microsoft.com/.default"],
        )

    def test_confidential_client_acquire_token_by_username_password(self):
        """Test ROPC flow with a confidential client."""
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        password = get_user_password(user)
        # Use a confidential client app that supports ROPC
        app = get_app_config(AppSecrets.S2S_CLIENT)
        client_secret = get_secret(app.secret_name, vault="msal_team")
        self._test_username_password(
            authority=user.authority,
            client_id=app.app_id,
            username=user.upn,
            password=password,
            scope=["https://graph.microsoft.com/.default"],
            client_secret=client_secret,
        )

    def _build_b2c_authority(self, policy):
        base = "https://msidlabb2c.b2clogin.com/msidlabb2c.onmicrosoft.com"
        return base + "/" + policy  # We do not support base + "?p=" + policy

    def manual_test_b2c_acquire_token_by_auth_code(self):
        """Test auth code flow for B2C."""
        # TODO: Update with B2C app config from Key Vault when available
        app = get_app_config(AppSecrets.B2C_CLIENT)  # TODO: Verify secret name
        self._test_acquire_token_by_auth_code(
            authority=self._build_b2c_authority("B2C_1_SignInPolicy"),
            client_id=app.app_id,
            port=3843,  # Lab defines 4 of them: [3843, 4584, 4843, 60000]
            scope=app.defaultscopes,
        )

    def manual_test_b2c_acquire_token_by_auth_code_flow(self):
        """Test auth code flow (with PKCE) for B2C."""
        user = get_user_config(UserSecrets.B2C)  # TODO: Verify secret name
        app = get_app_config(AppSecrets.B2C_CLIENT)  # TODO: Verify secret name
        self._test_acquire_token_by_auth_code_flow(
            authority=self._build_b2c_authority("B2C_1_SignInPolicy"),
            client_id=app.app_id,
            username=user.upn,
            lab_name=user.lab_name,
            port=3843,  # Lab defines 4 of them: [3843, 4584, 4843, 60000]
            scope=app.defaultscopes,
        )

    def test_b2c_acquire_token_by_ropc(self):
        """Test ROPC flow for B2C with custom authority."""
        user = get_user_config(UserSecrets.B2C)
        password = get_user_password(user)
        app = get_app_config(AppSecrets.B2C_CLIENT)
        # B2C uses a specific API scope, not generic app.scopes
        b2c_scope = ["https://msidlabb2c.onmicrosoft.com/msidlabb2capi/read"]
        self._test_username_password(
            authority=self._build_b2c_authority("B2C_1_ROPC_Auth"),
            client_id=app.app_id,
            username=user.upn,
            password=password,
            scope=b2c_scope,
        )

    def test_b2c_allows_using_client_id_as_scope(self):
        """Test that B2C allows using client_id as scope.
        
        See also https://learn.microsoft.com/en-us/azure/active-directory-b2c/access-tokens#openid-connect-scopes
        """
        user = get_user_config(UserSecrets.B2C)
        password = get_user_password(user)
        app = get_app_config(AppSecrets.B2C_CLIENT)
        self._test_username_password(
            authority=self._build_b2c_authority("B2C_1_ROPC_Auth"),
            client_id=app.app_id,
            username=user.upn,
            password=password,
            scope=[app.app_id],  # Using client_id as scope
        )


class CiamTestCase(LabBasedTestCase):
    """Test cases for CIAM (Customer Identity and Access Management).
    
    CIAM uses a different authority format: https://<labName>.ciamlogin.com/
    """

    @classmethod
    def setUpClass(cls):
        super(CiamTestCase, cls).setUpClass()
        # Get CIAM user and app config from Key Vault
        cls.user_config = get_user_config(UserSecrets.CIAM)
        cls.app_config = get_app_config(AppSecrets.CIAM_CLIENT)
        # CIAM authority format: https://<labName>.ciamlogin.com/
        cls.ciam_authority = f"https://{cls.user_config.lab_name}.ciamlogin.com/"

    def manual_test_ciam_acquire_token_interactive(self):
        """Test interactive flow for CIAM."""
        self._test_acquire_token_interactive(
            authority=self.ciam_authority,
            client_id=self.app_config.app_id,
            scope=self.app_config.defaultscopes,
            username=self.user_config.upn,
            lab_name=self.user_config.lab_name,
        )

    def test_ciam_acquire_token_for_client(self):
        """Test client credentials flow for CIAM using certificate."""
        self._test_acquire_token_by_client_cert(
            client_id=self.app_config.app_id,
            authority=self.ciam_authority,
            scope=[f"{self.app_config.app_id}/.default"],
        )

    def test_ciam_acquire_token_by_ropc(self):
        """Test ROPC flow for CIAM.
        
        Note: CIAM does not officially support ROPC for external emails,
        but this test works because the test data uses a local email.
        """
        password = get_user_password(self.user_config)
        self._test_username_password(
            authority=self.ciam_authority,
            client_id=self.app_config.app_id,
            username=self.user_config.upn,
            password=password,
            scope=["user.read"],
        )

    @unittest.skip("""As of Aug 2024, in both ciam2 and ciam6, sign-in fails with
AADSTS500208: The domain is not a valid login domain for the account type.""")
    def manual_test_ciam_device_flow(self):
        self._test_device_flow(
            authority=self.ciam_authority,
            client_id=self.app_config.app_id,
            scope=self.app_config.scopes or ["user.read"],
        )


class CiamCudTestCase(CiamTestCase):
    """CIAM CUD (Custom URL Domain) test case.
    
    Uses OIDC authority instead of standard CIAM authority for custom domain scenarios.
    """
    # OIDC authority for CIAM CUD
    CIAM_CUD_OIDC_AUTHORITY = "https://login.msidlabsciam.com/fe362aec-5d43-45d1-b730-9755e60dc3b9/v2.0"

    @classmethod
    def setUpClass(cls):
        super(CiamCudTestCase, cls).setUpClass()
        # Override authority to use OIDC authority for CUD scenario
        cls.ciam_authority = None  # Clear standard authority
        cls.oidc_authority = cls.CIAM_CUD_OIDC_AUTHORITY

    def manual_test_ciam_acquire_token_interactive(self):
        """Test interactive flow for CIAM CUD with OIDC authority."""
        self._test_acquire_token_interactive(
            oidc_authority=self.oidc_authority,
            client_id=self.app_config.app_id,
            scope=self.app_config.defaultscopes,
            username=self.user_config.upn,
            lab_name=self.user_config.lab_name,
        )

    def test_ciam_acquire_token_for_client(self):
        """Test client credentials flow for CIAM CUD using certificate with OIDC authority."""
        self._test_acquire_token_by_client_cert(
            client_id=self.app_config.app_id,
            oidc_authority=self.oidc_authority,
            scope=[".default"],
        )

    def test_ciam_acquire_token_by_ropc(self):
        """Test ROPC flow for CIAM CUD with OIDC authority."""
        password = get_user_password(self.user_config)
        self._test_username_password(
            oidc_authority=self.oidc_authority,
            client_id=self.app_config.app_id,
            username=self.user_config.upn,
            password=password,
            scope=["user.read"],
        )


class WorldWideRegionalEndpointTestCase(LabBasedTestCase):
    """Test regional endpoint behavior for confidential client applications.
    
    Regional endpoints are used for Azure-hosted applications to reduce latency.
    These tests verify that MSAL correctly routes requests to regional vs global endpoints.
    """
    region = "westus"
    timeout = 5  # Short timeout makes this test case responsive on non-VM

    def _test_acquire_token_for_client(self, configured_region, expected_region):
        """This is the only grant supported by regional endpoint, for now.
        
        Uses the lab app certificate for authentication.
        """
        from tests.lab_config import get_client_certificate
        
        # Get client ID from lab_config constant and certificate from lab_config
        if not clean_env("LAB_APP_CLIENT_CERT_PFX_PATH"):
            self.skipTest("LAB_APP_CLIENT_CERT_PFX_PATH is required")
        client_credential = get_client_certificate()
        
        self.app = msal.ConfidentialClientApplication(
            LAB_APP_CLIENT_ID,
            client_credential=client_credential,
            authority="https://login.microsoftonline.com/microsoft.onmicrosoft.com",
            azure_region=configured_region,
            http_client=MinimalHttpClient(timeout=self.timeout),
        )
        scopes = ["https://graph.microsoft.com/.default"]

        with patch.object(  # Test the request hit the regional endpoint
                self.app.http_client, "post", return_value=MinimalResponse(
                status_code=400, text='{"error": "mock"}')) as mocked_method:
            self.app.acquire_token_for_client(scopes)
            expected_host = '{}.login.microsoft.com'.format(
                expected_region) if expected_region else 'login.microsoftonline.com'
            mocked_method.assert_called_with(
                'https://{}/{}/oauth2/v2.0/token'.format(
                    expected_host, self.app.authority.tenant),
                params=ANY, data=ANY, headers=ANY)
        result = self.app.acquire_token_for_client(
            scopes,
            params={"AllowEstsRNonMsi": "true"},  # For testing regional endpoint. It will be removed once MSAL Python 1.12+ has been onboard to ESTS-R
            )
        self.assertIn('access_token', result)
        self.assertCacheWorksForApp(result, scopes)

    def test_acquire_token_for_client_should_hit_global_endpoint_by_default(self):
        self._test_acquire_token_for_client(None, None)

    def test_acquire_token_for_client_should_ignore_env_var_region_name_by_default(self):
        os.environ["REGION_NAME"] = "eastus"
        self._test_acquire_token_for_client(None, None)
        del os.environ["REGION_NAME"]

    @patch.dict(os.environ, {"MSAL_FORCE_REGION": "eastus"})
    def test_acquire_token_for_client_should_use_env_var_msal_force_region_by_default(self):
        self._test_acquire_token_for_client(None, "eastus")

    @patch.dict(os.environ, {"MSAL_FORCE_REGION": "eastus"})
    def test_acquire_token_for_client_should_prefer_the_explicit_region(self):
        self._test_acquire_token_for_client("westus", "westus")

    @patch.dict(os.environ, {"MSAL_FORCE_REGION": "eastus"})
    def test_acquire_token_for_client_should_allow_opt_out_env_var_msal_force_region(self):
        self._test_acquire_token_for_client(False, None)

    def test_acquire_token_for_client_should_use_a_specified_region(self):
        self._test_acquire_token_for_client("westus", "westus")

    def test_acquire_token_for_client_should_use_an_env_var_with_short_region_name(self):
        os.environ["REGION_NAME"] = "eastus"
        self._test_acquire_token_for_client(
            msal.ConfidentialClientApplication.ATTEMPT_REGION_DISCOVERY, "eastus")
        del os.environ["REGION_NAME"]

    def test_acquire_token_for_client_should_use_an_env_var_with_long_region_name(self):
        os.environ["REGION_NAME"] = "East Us 2"
        self._test_acquire_token_for_client(
            msal.ConfidentialClientApplication.ATTEMPT_REGION_DISCOVERY, "eastus2")
        del os.environ["REGION_NAME"]

    def test_cca_obo_should_bypass_regional_endpoint_therefore_still_work(self):
        """We test OBO because it is implemented in sub class ConfidentialClientApplication.
        
        Regional endpoint does not directly support OBO, but MSAL should automatically
        bypass regional endpoint for OBO calls.
        """
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        password = get_user_password(user)
        web_api_app = get_app_config(AppSecrets.WEB_API_CLIENT)

        # Step 1: PCA gets token for user to access the WebAPI
        config_pca = {
            "authority": "https://login.microsoftonline.com/organizations",
            "client_id": web_api_app.app_id,
            "username": user.upn,
            "password": password,
            "scope": ["api://%s/access_as_user" % web_api_app.app_id],
        }

        # Step 2: WebAPI (CCA) exchanges the token via OBO for Graph access
        config_cca = {
            "authority": user.authority,
            "client_id": web_api_app.app_id,
            "client_secret": get_secret(web_api_app.client_secret, vault="msal_team"),
            "scope": ["https://graph.microsoft.com/.default"],
            "username": user.upn,
        }

        self._test_acquire_token_obo(
            config_pca, config_cca,
            azure_region=self.region,
            http_client=MinimalHttpClient(timeout=self.timeout),
        )

    def test_cca_ropc_should_bypass_regional_endpoint_therefore_still_work(self):
        """We test ROPC because it is implemented in base class ClientApplication.
        
        Regional endpoint does not directly support ROPC, but MSAL should automatically
        bypass regional endpoint for ROPC calls.
        """
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        password = get_user_password(user)
        # Use the S2S app which supports ROPC with client secret
        app = get_app_config(AppSecrets.S2S_CLIENT)
        client_secret = get_secret(app.secret_name, vault="msal_team")
        
        self._test_username_password(
            authority=user.authority,
            client_id=app.app_id,
            username=user.upn,
            password=password,
            scope=["https://graph.microsoft.com/.default"],
            client_secret=client_secret,
            azure_region=self.region,
            http_client=MinimalHttpClient(timeout=self.timeout),
        )


class ArlingtonCloudTestCase(LabBasedTestCase):
    """Test cases for Azure US Government (Arlington) cloud.
    
    Arlington uses different endpoints (login.microsoftonline.us, graph.microsoft.us)
    and requires specific test resources configured for the government cloud.
    """

    @classmethod
    def setUpClass(cls):
        super(ArlingtonCloudTestCase, cls).setUpClass()
        # Get Arlington user and app config from Key Vault
        cls.user_config = get_user_config(UserSecrets.ARLINGTON)
        cls.app_config = get_app_config(AppSecrets.ARLINGTON_CLIENT)
        # Arlington authority uses login.microsoftonline.us
        cls.arlington_authority = f"https://login.microsoftonline.us/{cls.user_config.tenant_id}"

    def test_acquire_token_by_ropc(self):
        """Test username/password flow for Arlington cloud."""
        password = get_user_password(self.user_config)
        self._test_username_password(
            authority=self.arlington_authority,
            client_id=self.app_config.app_id,
            username=self.user_config.upn,
            password=password,
            scope=["https://graph.microsoft.us/.default"],
        )

    def manual_test_acquire_token_device_flow(self):
        """Test device code flow for Arlington cloud."""
        self._test_device_flow(
            authority=self.arlington_authority,
            client_id=self.app_config.app_id,
            scope=["user.read"],
            username=self.user_config.upn,
            lab_name=self.user_config.lab_name,
        )

    def test_acquire_token_silent_with_an_empty_cache_should_return_none(self):
        """Test that acquire_token_silent returns None with empty cache.
        
        Note: An alias in this region is no longer accepting HTTPS traffic.
        If this test case passes without exception, it means MSAL Python
        is not affected by that.
        """
        client_secret = get_secret(self.app_config.client_secret, vault="msal_team")
        app = msal.ConfidentialClientApplication(
            self.app_config.app_id,
            client_credential=client_secret,
            authority=self.arlington_authority,
            http_client=MinimalHttpClient(),
        )
        result = app.acquire_token_silent(
            scopes=["https://graph.microsoft.us/.default"],
            account=None,
        )
        self.assertEqual(result, None)


@unittest.skipUnless(_PYMSALRUNTIME_INSTALLED, "AT POP feature is only supported by using broker")
class PopTestCase(LabBasedTestCase):
    """Test cases for Proof-of-Possession (POP) tokens using the broker.
    
    These tests require pymsalruntime to be installed.
    """

    def manual_test_at_pop_should_contain_pop_scheme_content(self):
        auth_scheme = msal.PopAuthScheme(
            http_method=msal.PopAuthScheme.HTTP_GET,
            url="https://www.Contoso.com/Path1/Path2?queryParam1=a&queryParam2=b",
            nonce="placeholder",
            )
        result = self._test_acquire_token_interactive(
            # Lab test users tend to get kicked out from WAM, we use local user to test
            client_id=_AZURE_CLI,
            authority="https://login.microsoftonline.com/organizations",
            scope=["https://management.azure.com/.default"],
            auth_scheme=auth_scheme,
            )   # It also tests assertCacheWorksForUser()
        self.assertEqual(result["token_source"], "broker", "POP is only supported by broker")
        self.assertEqual(result["token_type"], "pop")
        payload = json.loads(decode_part(result["access_token"].split(".")[1]))
        logger.debug("AT POP payload = %s", json.dumps(payload, indent=2))
        self.assertEqual(payload["m"], auth_scheme._http_method)
        self.assertEqual(payload["u"], auth_scheme._url.netloc)
        self.assertEqual(payload["p"], auth_scheme._url.path)
        self.assertEqual(payload["nonce"], auth_scheme._nonce)

    # TODO: Remove this, as ROPC support is removed by Broker-on-Win
    def manual_test_at_pop_via_testingsts_service(self):
        """Test AT POP via testingsts.azurewebsites.net nonce validation service."""
        self.skipTest("ROPC support is removed by Broker-on-Win")
        auth_scheme = msal.PopAuthScheme(
            http_method="POST",
            url="https://www.Contoso.com/Path1/Path2?queryParam1=a&queryParam2=b",
            nonce=requests.get(
                # TODO: Could use ".../missing" and then parse its WWW-Authenticate header
                "https://testingsts.azurewebsites.net/servernonce/get").text,
            )
        # Use Key Vault config for user and app
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        password = get_user_password(user)
        app = get_app_config(AppSecrets.PCA_CLIENT)
        result = self._test_username_password(
            authority=user.authority,
            client_id=app.app_id,
            username=user.upn,
            password=password,
            scope=["User.Read"],
            auth_scheme=auth_scheme,
        )
        self.assertEqual(result["token_type"], "pop")
        shr = result["access_token"]
        payload = json.loads(decode_part(result["access_token"].split(".")[1]))
        logger.debug("AT POP payload = %s", json.dumps(payload, indent=2))
        self.assertEqual(payload["m"], auth_scheme._http_method)
        self.assertEqual(payload["u"], auth_scheme._url.netloc)
        self.assertEqual(payload["p"], auth_scheme._url.path)
        self.assertEqual(payload["nonce"], auth_scheme._nonce)

        validation = requests.post(
            # TODO: This endpoint does not seem to validate the url
            "https://testingsts.azurewebsites.net/servernonce/validateshr",
            data={"SHR": shr},
            )
        self.assertEqual(validation.status_code, 200)

    def manual_test_at_pop_calling_pattern(self):
        """Test the POP calling pattern with Key Vault config.
        
        The calling pattern was described here:
        https://identitydivision.visualstudio.com/DevEx/_git/AuthLibrariesApiReview?path=/PoPTokensProtocol/PoP_API_In_MSAL.md
        """
        # It is supposed to call app.is_pop_supported() first,
        # and then fallback to bearer token code path.
        # We skip it here because this test case has not yet initialize self.app
        # assert self.app.is_pop_supported()

        api_endpoint = "https://20.190.132.47/beta/me"
        verify = True  # Hopefully this will make CodeQL happy
        if verify:
            self.skipTest("""
            The api_endpoint is for test only and has no proper SSL certificate,
            so you would have to disable SSL certificate checks and run this test case manually.
            We tried suppressing the CodeQL warning by adding this in the proper places
                @suppress py/bandit/requests-ssl-verify-disabled
            but it did not work.
            """)
        # @suppress py/bandit/requests-ssl-verify-disabled
        resp = requests.get(api_endpoint, verify=verify)  # CodeQL [SM03157]
        self.assertEqual(resp.status_code, 401, "Initial call should end with an http 401 error")
        
        # Use Key Vault config instead of Lab API
        user = get_user_config(UserSecrets.PUBLIC_CLOUD)
        result = self._get_shr_pop(
            client_id=get_app_config(AppSecrets.PCA_CLIENT).app_id,
            authority=user.authority,
            scope=["https://graph.microsoft.com/.default"],
            username=user.upn,
            lab_name=user.lab_name,
            auth_scheme=msal.PopAuthScheme(
                http_method=msal.PopAuthScheme.HTTP_GET,
                url=api_endpoint,
                nonce=self._extract_pop_nonce(resp.headers.get("WWW-Authenticate")),
            ),
        )
        resp = requests.get(
            api_endpoint,
            # CodeQL [SM03157]
            verify=verify,  # @suppress py/bandit/requests-ssl-verify-disabled
            headers={
            "Authorization": "pop {}".format(result["access_token"]),
            })
        self.assertEqual(resp.status_code, 200, "POP resource should be accessible")

    def _extract_pop_nonce(self, www_authenticate):
        # This is a hack for testing purpose only. Do not use this in prod.
        # FYI: There is a www-authenticate package but it falters when encountering realm=""
        import re
        found = re.search(r'nonce="(.+?)"', www_authenticate)
        if found:
            return found.group(1)

    def _get_shr_pop(
            self, client_id=None, authority=None, scope=None, auth_scheme=None,
            **kwargs):
        result = self._test_acquire_token_interactive(
            # Lab test users tend to get kicked out from WAM, we use local user to test
            client_id=client_id,
            authority=authority,
            scope=scope,
            auth_scheme=auth_scheme,
            **kwargs)  # It also tests assertCacheWorksForUser()
        self.assertEqual(result["token_source"], "broker", "POP is only supported by broker")
        self.assertEqual(result["token_type"], "pop")
        payload = json.loads(decode_part(result["access_token"].split(".")[1]))
        logger.debug("AT POP payload = %s", json.dumps(payload, indent=2))
        self.assertEqual(payload["m"], auth_scheme._http_method)
        self.assertEqual(payload["u"], auth_scheme._url.netloc)
        self.assertEqual(payload["p"], auth_scheme._url.path)
        self.assertEqual(payload["nonce"], auth_scheme._nonce)
        return result


if __name__ == "__main__":
    unittest.main()



import unittest
import socket
import sys

import requests

from msal.oauth2cli.authcode import AuthCodeReceiver


class TestAuthCodeReceiver(unittest.TestCase):
    def test_setup_at_a_given_port_and_teardown(self):
        port = 12345  # Assuming this port is available
        with AuthCodeReceiver(port=port) as receiver:
            self.assertEqual(port, receiver.get_port())

    def test_setup_at_a_ephemeral_port_and_teardown(self):
        port = 0
        with AuthCodeReceiver(port=port) as receiver:
            self.assertNotEqual(port, receiver.get_port())

    def test_no_two_concurrent_receivers_can_listen_on_same_port(self):
        with AuthCodeReceiver() as receiver:
            expected_error = OSError if sys.version_info[0] > 2 else socket.error
            with self.assertRaises(expected_error):
                with AuthCodeReceiver(port=receiver.get_port()):
                    pass

    def test_template_should_escape_input(self):
        """Test that HTML in error response is properly escaped"""
        with AuthCodeReceiver() as receiver:
            receiver._scheduled_actions = [(  # Injection happens here when the port is known
                1,  # Delay it until the receiver is activated by get_auth_response()
                lambda: self.assertEqual(
                    "<html>&lt;script&gt;alert(&#x27;xss&#x27;);&lt;/script&gt;</html>",
                    requests.post(
                        "http://localhost:{}".format(receiver.get_port()),
                        data={"error": "<script>alert('xss');</script>"},
                    ).text,
            ))]
            receiver.get_auth_response(  # Starts server and hang until timeout
                timeout=3,
                error_template="<html>$error</html>",
            )

    def test_get_request_with_auth_code_is_rejected(self):
        """Test that GET request with auth code is rejected for security"""
        with AuthCodeReceiver() as receiver:
            test_state = "test_state_67890"
            receiver._scheduled_actions = [(
                1,
                lambda: self.assertEqual(400, requests.get(
                    "http://localhost:{}".format(receiver.get_port()), params={
                        "code": "test_auth_code_12345",
                        "state": test_state
                    }
                ).status_code)
            )]
            result = receiver.get_auth_response(timeout=3, state=test_state)
            self.assertIsNone(result, "Should not receive auth response via GET")

    def test_post_request_with_auth_code(self):
        """Test that POST request with auth code is handled correctly (form_post response mode)"""
        with AuthCodeReceiver() as receiver:
            test_code = "test_auth_code_12345"
            test_state = "test_state_67890"
            receiver._scheduled_actions = [(
                1,
                lambda: requests.post(
                    "http://localhost:{}".format(receiver.get_port()),
                    data={"code": test_code, "state": test_state},
                )
            )]
            result = receiver.get_auth_response(timeout=3, state=test_state)
            self.assertIsNotNone(result, "Should receive auth response via POST")
            self.assertEqual(result.get("code"), test_code)
            self.assertEqual(result.get("state"), test_state)

    def test_post_request_with_error(self):
        """Test that POST request with error is handled correctly"""
        with AuthCodeReceiver() as receiver:
            test_error = "access_denied"
            test_error_description = "User denied access"
            receiver._scheduled_actions = [(
                1,
                lambda: requests.post(
                    "http://localhost:{}".format(receiver.get_port()),
                    data={"error": test_error, "error_description": test_error_description},
                )
            )]
            result = receiver.get_auth_response(timeout=3)
            self.assertIsNotNone(result, "Should receive auth response via POST")
            self.assertEqual(result.get("error"), test_error)
            self.assertEqual(result.get("error_description"), test_error_description)

    def test_post_request_state_mismatch(self):
        """Test that POST request with mismatched state is rejected"""
        with AuthCodeReceiver() as receiver:
            receiver._scheduled_actions = [(
                1,
                lambda: requests.post(
                    "http://localhost:{}".format(receiver.get_port()),
                    data={"code": "test_code", "state": "wrong_state"},
                )
            )]
            result = receiver.get_auth_response(timeout=3, state="expected_state")
            self.assertIsNone(result, "Should not receive auth response due to state mismatch")



# Test cases for https://identitydivision.visualstudio.com/devex/_git/AuthLibrariesApiReview?version=GBdev&path=%2FService%20protection%2FIntial%20set%20of%20protection%20measures.md&_a=preview&anchor=common-test-cases
import pickle
from time import sleep
from random import random
import logging

from msal.throttled_http_client import (
    ThrottledHttpClientBase, ThrottledHttpClient, NormalizedResponse)
from msal.exceptions import MsalServiceError

from tests import unittest
from tests.http_client import MinimalResponse as _MinimalResponse


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class MinimalResponse(_MinimalResponse):
    SIGNATURE = str(random()).encode("utf-8")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ = (  # Only an instance attribute will be stored in pickled instance
            self.__class__.SIGNATURE)  # Useful for testing its presence in pickled instance


class DummyHttpClient(object):
    def __init__(self, status_code=None, response_headers=None, response_text=None):
        self._status_code = status_code
        self._response_headers = response_headers
        self._response_text = response_text

    def _build_dummy_response(self):
        return MinimalResponse(
            status_code=self._status_code,
            headers=self._response_headers,
            text=self._response_text if self._response_text is not None else str(
                random()  # So that we'd know whether a new response is received
            ),
        )

    def post(self, url, params=None, data=None, headers=None, **kwargs):
        return self._build_dummy_response()

    def get(self, url, params=None, headers=None, **kwargs):
        return self._build_dummy_response()

    def close(self):
        raise CloseMethodCalled("Not used by MSAL, but our customers may use it")


class CloseMethodCalled(Exception):
    pass


class NormalizedResponseTestCase(unittest.TestCase):
    def test_pickled_minimal_response_should_contain_signature(self):
        self.assertIn(MinimalResponse.SIGNATURE, pickle.dumps(MinimalResponse(
            status_code=200, headers={}, text="foo")))

    def test_normalized_response_should_not_contain_signature(self):
        response = NormalizedResponse(MinimalResponse(
            status_code=200, headers={}, text="foo"))
        self.assertNotIn(
            MinimalResponse.SIGNATURE, pickle.dumps(response),
            "A pickled object should not contain undesirable data")
        self.assertEqual(response.text, "foo", "Should return the same response text")

    def test_normalized_response_raise_for_status_should_raise(self):
        response = NormalizedResponse(MinimalResponse(
            status_code=400, headers={}, text="foo"))
        with self.assertRaises(MsalServiceError):
            response.raise_for_status()


class ThrottledHttpClientBaseTestCase(unittest.TestCase):

    def assertCleanPickle(self, obj):
        self.assertTrue(bool(obj), "The object should not be empty")
        self.assertNotIn(
            MinimalResponse.SIGNATURE, pickle.dumps(obj),
            "A pickled object should not contain undesirable data")

    def assertValidResponse(self, response):
        self.assertIsInstance(response, NormalizedResponse)
        self.assertCleanPickle(response)

    def test_throttled_http_client_base_response_should_tolerate_headerless_response(self):
        # MSAL Python 1.32.1 had a regression that caused it to require headers in the response.
        # This was fixed in 1.32.2
        # https://github.com/AzureAD/microsoft-authentication-library-for-python/compare/1.32.1...1.32.2
        # This test case is to ensure that we can tolerate headerless response.
        http_client = DummyHttpClient(status_code=200, response_text="foo")
        raw_response = http_client.post("https://example.com")
        self.assertFalse(hasattr(raw_response, "headers"), "Should not contain headers")
        response = ThrottledHttpClientBase(http_client).post("https://example.com")
        self.assertEqual(response.text, "foo", "Should return the same response text")

    def test_throttled_http_client_base_response_should_not_contain_signature(self):
        http_client = ThrottledHttpClientBase(DummyHttpClient(status_code=200))
        response = http_client.post("https://example.com")
        self.assertValidResponse(response)

    def assertNotAlteringOriginalHttpClient(self, ThrottledHttpClientClass):
        original_http_client = DummyHttpClient()
        original_get = original_http_client.get
        original_post = original_http_client.post
        throttled_http_client = ThrottledHttpClientClass(original_http_client)
        goal = """The implementation should wrap original http_client
            and keep it intact, instead of monkey-patching it"""
        self.assertNotEqual(throttled_http_client, original_http_client, goal)
        self.assertEqual(original_post, original_http_client.post)
        self.assertEqual(original_get, original_http_client.get)

    def test_throttled_http_client_base_should_not_alter_original_http_client(self):
        self.assertNotAlteringOriginalHttpClient(ThrottledHttpClientBase)

    def test_throttled_http_client_base_should_not_nest_http_client(self):
        original_http_client = DummyHttpClient()
        throttled_http_client = ThrottledHttpClientBase(original_http_client)
        self.assertIs(original_http_client, throttled_http_client.http_client)
        nested_throttled_http_client = ThrottledHttpClientBase(throttled_http_client)
        self.assertIs(original_http_client, nested_throttled_http_client.http_client)


class ThrottledHttpClientTestCase(ThrottledHttpClientBaseTestCase):

    def test_throttled_http_client_should_not_alter_original_http_client(self):
        self.assertNotAlteringOriginalHttpClient(ThrottledHttpClient)

    def _test_RetryAfter_N_seconds_should_keep_entry_for_N_seconds(
            self, http_client, retry_after):
        http_cache = {}
        http_client = ThrottledHttpClient(http_client, http_cache=http_cache)
        resp1 = http_client.post("https://example.com")  # We implemented POST only
        resp2 = http_client.post("https://example.com")  # We implemented POST only
        logger.debug(http_cache)
        self.assertEqual(resp1.text, resp2.text, "Should return a cached response")
        sleep(retry_after + 1)
        resp3 = http_client.post("https://example.com")  # We implemented POST only
        self.assertNotEqual(resp1.text, resp3.text, "Should return a new response")

    def test_429_with_RetryAfter_N_seconds_should_keep_entry_for_N_seconds(self):
        retry_after = 1
        self._test_RetryAfter_N_seconds_should_keep_entry_for_N_seconds(
            DummyHttpClient(
                status_code=429, response_headers={"Retry-After": retry_after}),
            retry_after)

    def test_5xx_with_RetryAfter_N_seconds_should_keep_entry_for_N_seconds(self):
        retry_after = 1
        self._test_RetryAfter_N_seconds_should_keep_entry_for_N_seconds(
            DummyHttpClient(
                status_code=503, response_headers={"Retry-After": retry_after}),
            retry_after)

    def test_400_with_RetryAfter_N_seconds_should_keep_entry_for_N_seconds(self):
        """Retry-After is supposed to only shown in http 429/5xx,
        but we choose to support Retry-After for arbitrary http response."""
        retry_after = 1
        self._test_RetryAfter_N_seconds_should_keep_entry_for_N_seconds(
            DummyHttpClient(
                status_code=400, response_headers={"Retry-After": retry_after}),
            retry_after)

    def test_one_RetryAfter_request_should_block_a_similar_request(self):
        http_cache = {}
        http_client = DummyHttpClient(
            status_code=429, response_headers={"Retry-After": 2})
        http_client = ThrottledHttpClient(http_client, http_cache=http_cache)
        resp1 = http_client.post("https://example.com", data={
            "scope": "one", "claims": "bar", "grant_type": "authorization_code"})
        resp2 = http_client.post("https://example.com", data={
            "scope": "one", "claims": "foo", "grant_type": "password"})
        logger.debug(http_cache)
        self.assertEqual(resp1.text, resp2.text, "Should return a cached response")

    def test_one_RetryAfter_request_should_not_block_a_different_request(self):
        http_cache = {}
        http_client = DummyHttpClient(
            status_code=429, response_headers={"Retry-After": 2})
        http_client = ThrottledHttpClient(http_client, http_cache=http_cache)
        resp1 = http_client.post("https://example.com", data={"scope": "one"})
        resp2 = http_client.post("https://example.com", data={"scope": "two"})
        logger.debug(http_cache)
        self.assertNotEqual(resp1.text, resp2.text, "Should return a new response")

    def test_one_invalid_grant_should_block_a_similar_request(self):
        http_cache = {}
        http_client = DummyHttpClient(
            status_code=400)  # It covers invalid_grant and interaction_required
        http_client = ThrottledHttpClient(http_client, http_cache=http_cache)

        resp1 = http_client.post("https://example.com", data={"claims": "foo"})
        logger.debug(http_cache)
        self.assertValidResponse(resp1)
        resp1_again = http_client.post("https://example.com", data={"claims": "foo"})
        self.assertValidResponse(resp1_again)
        self.assertEqual(resp1.text, resp1_again.text, "Should return a cached response")

        resp2 = http_client.post("https://example.com", data={"claims": "bar"})
        self.assertValidResponse(resp2)
        self.assertNotEqual(resp1.text, resp2.text, "Should return a new response")
        resp2_again = http_client.post("https://example.com", data={"claims": "bar"})
        self.assertValidResponse(resp2_again)
        self.assertEqual(resp2.text, resp2_again.text, "Should return a cached response")

        self.assertCleanPickle(http_cache)

    def test_one_foci_app_recovering_from_invalid_grant_should_also_unblock_another(self):
        """
        Need not test multiple FOCI app's acquire_token_silent() here. By design,
        one FOCI app's successful populating token cache would result in another
        FOCI app's acquire_token_silent() to hit a token without invoking http request.
        """

    def test_forcefresh_behavior(self):
        """
        The implementation let token cache and http cache operate in different
        layers. They do not couple with each other.
        Therefore, acquire_token_silent(..., force_refresh=True)
        would bypass the token cache yet technically still hit the http cache.

        But that is OK, cause the customer need no force_refresh in the first place.
        After a successful AT/RT acquisition, AT/RT will be in the token cache,
        and a normal acquire_token_silent(...) without force_refresh would just work.
        This was discussed in https://identitydivision.visualstudio.com/DevEx/_git/AuthLibrariesApiReview/pullrequest/3618?_a=files
        """

    def test_http_get_200_should_be_cached(self):
        http_cache = {}
        http_client = DummyHttpClient(
            status_code=200)  # It covers UserRealm discovery and OIDC discovery
        http_client = ThrottledHttpClient(http_client, http_cache=http_cache)
        resp1 = http_client.get("https://example.com?foo=bar")
        resp2 = http_client.get("https://example.com?foo=bar")
        logger.debug(http_cache)
        self.assertEqual(resp1.text, resp2.text, "Should return a cached response")

    def test_device_flow_retry_should_not_be_cached(self):
        DEVICE_AUTH_GRANT = "urn:ietf:params:oauth:grant-type:device_code"
        http_cache = {}
        http_client = DummyHttpClient(status_code=400)
        http_client = ThrottledHttpClient(http_client, http_cache=http_cache)
        resp1 = http_client.post(
            "https://example.com", data={"grant_type": DEVICE_AUTH_GRANT})
        resp2 = http_client.post(
            "https://example.com", data={"grant_type": DEVICE_AUTH_GRANT})
        logger.debug(http_cache)
        self.assertNotEqual(resp1.text, resp2.text, "Should return a new response")

    def test_throttled_http_client_should_provide_close(self):
        http_client = DummyHttpClient(status_code=200)
        http_client = ThrottledHttpClient(http_client)
        with self.assertRaises(CloseMethodCalled):
            http_client.close()




import unittest
try:
    from unittest.mock import patch, ANY
except:
    from mock import patch, ANY

from tests.http_client import MinimalResponse
from tests.test_token_cache import build_response

import msal


class TestCcsRoutingInfoTestCase(unittest.TestCase):

    def test_acquire_token_by_auth_code_flow(self):
        app = msal.ClientApplication("client_id")
        state = "foo"
        flow = app.initiate_auth_code_flow(
            ["some", "scope"], login_hint="johndoe@contoso.com", state=state)
        with patch.object(app.http_client, "post", return_value=MinimalResponse(
                status_code=400, text='{"error": "mock"}')) as mocked_method:
            app.acquire_token_by_auth_code_flow(flow, {
                "state": state,
                "code": "bar",
                "client_info":  # MSAL asks for client_info, so it would be available
                    "eyJ1aWQiOiJhYTkwNTk0OS1hMmI4LTRlMGEtOGFlYS1iMzJlNTNjY2RiNDEiLCJ1dGlkIjoiNzJmOTg4YmYtODZmMS00MWFmLTkxYWItMmQ3Y2QwMTFkYjQ3In0",
                })
            self.assertEqual(
                "Oid:aa905949-a2b8-4e0a-8aea-b32e53ccdb41@72f988bf-86f1-41af-91ab-2d7cd011db47",
                mocked_method.call_args[1].get("headers", {}).get('X-AnchorMailbox'),
                "CSS routing info should be derived from client_info")

    # I've manually tested acquire_token_interactive. No need to automate it,
    # because it and acquire_token_by_auth_code_flow() share same code path.

    def test_acquire_token_silent(self):
        uid = "foo"
        utid = "bar"
        client_id = "my_client_id"
        scopes = ["some", "scope"]
        authority_url = "https://login.microsoftonline.com/common"
        token_cache = msal.TokenCache()
        token_cache.add({  # Pre-populate the cache
            "client_id": client_id,
            "scope": scopes,
            "token_endpoint": "{}/oauth2/v2.0/token".format(authority_url),
            "response": build_response(
                access_token="an expired AT to trigger refresh", expires_in=-99,
                uid=uid, utid=utid, refresh_token="this is a RT"),
            })  # The add(...) helper populates correct home_account_id for future searching
        app = msal.ClientApplication(
            client_id, authority=authority_url, token_cache=token_cache)
        with patch.object(app.http_client, "post", return_value=MinimalResponse(
                status_code=400, text='{"error": "mock"}')) as mocked_method:
            account = {"home_account_id": "{}.{}".format(uid, utid)}
            app.acquire_token_silent(["scope"], account)
            self.assertEqual(
                "Oid:{}@{}".format(  # Server accepts case-insensitive value
                    uid, utid),  # It would look like "Oid:foo@bar"
                mocked_method.call_args[1].get("headers", {}).get('X-AnchorMailbox'),
                "CSS routing info should be derived from home_account_id")

    def test_acquire_token_by_username_password(self):
        import warnings
        app = msal.ClientApplication("client_id")
        username = "johndoe@contoso.com"
        with patch.object(app.http_client, "post", return_value=MinimalResponse(
                status_code=400, text='{"error": "mock"}')) as mocked_method:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", DeprecationWarning)
                app.acquire_token_by_username_password(username, "password", ["scope"])
            self.assertEqual(
                "upn:" + username,
                mocked_method.call_args[1].get("headers", {}).get('X-AnchorMailbox'),
                "CSS routing info should be derived from client_info")




from functools import wraps
import time
try:
    from collections.abc import MutableMapping  # Python 3.3+
except ImportError:
    from collections import MutableMapping  # Python 2.7+
import heapq
from threading import Lock


class _ExpiringMapping(MutableMapping):
    _INDEX = "_index_"

    def __init__(self, mapping=None, capacity=None, expires_in=None, lock=None,
        *args, **kwargs):
        """Items in this mapping can have individual shelf life,
        just like food items in your refrigerator have their different shelf life
        determined by each food, not by the refrigerator.

        Expired items will be automatically evicted.
        The clean-up will be done at each time when adding a new item,
        or when looping or counting the entire mapping.
        (This is better than being done indecisively by a background thread,
        which might not always happen before your accessing the mapping.)

        This implementation uses no dependency other than Python standard library.

        :param MutableMapping mapping:
            A dict-like key-value mapping, which needs to support __setitem__(),
            __getitem__(), __delitem__(), get(), pop().

            The default mapping is an in-memory dict.

            You could potentially supply a file-based dict-like object, too.
            This implementation deliberately avoid mapping.__iter__(),
            which could be slow on a file-based mapping.

        :param int capacity:
            How many items this mapping will hold.
            When you attempt to add new item into a full mapping,
            it will automatically delete the item that is expiring soonest.

            The default value is None, which means there is no capacity limit.

        :param int expires_in:
            How many seconds an item would expire and be purged from this mapping.
            Also known as time-to-live (TTL).
            You can also use :func:`~set()` to provide per-item expires_in value.

        :param Lock lock:
            A locking mechanism with context manager interface.
            If no lock is provided, a threading.Lock will be used.
            But you may want to supply a different lock,
            if your customized mapping is being shared differently.
        """
        super(_ExpiringMapping, self).__init__(*args, **kwargs)
        self._mapping = mapping if mapping is not None else {}
        self._capacity = capacity
        self._expires_in = expires_in
        self._lock = Lock() if lock is None else lock

    def _peek(self):
        # Returns (sequence, timestamps) without triggering maintenance
        return self._mapping.get(self._INDEX, ([], {}))

    def _validate_key(self, key):
        if key == self._INDEX:
            raise ValueError("key {} is a reserved keyword in {}".format(
                key, self.__class__.__name__))

    def set(self, key, value, expires_in):
        # This method's name was chosen so that it matches its cousin __setitem__(),
        # and it also complements the counterpart get().
        # The downside is such a name shadows the built-in type set in this file,
        # but you can overcome that by defining a global alias for set.
        """It sets the key-value pair into this mapping, with its per-item expires_in.

        It will take O(logN) time, because it will run some maintenance.
        This worse-than-constant time is acceptable, because in a cache scenario,
        __setitem__() would only be called during a cache miss,
        which would already incur an expensive target function call anyway.

        By the way, most other methods of this mapping still have O(1) constant time.
        """
        with self._lock:
            self._set(key, value, expires_in)

    def _set(self, key, value, expires_in):
        # This internal implementation powers both set() and __setitem__(),
        # so that they don't depend on each other.
        self._validate_key(key)
        sequence, timestamps = self._peek()
        self._maintenance(sequence, timestamps)  # O(logN)
        now = int(time.time())
        expires_at = now + expires_in
        entry = [expires_at, now, key]
        is_new_item = key not in timestamps
        is_beyond_capacity = self._capacity and len(timestamps) >= self._capacity
        if is_new_item and is_beyond_capacity:
            self._drop_indexed_entry(timestamps, heapq.heappushpop(sequence, entry))
        else:  # Simply add new entry. The old one would become a harmless orphan.
            heapq.heappush(sequence, entry)
        timestamps[key] = [expires_at, now]  # It overwrites existing key, if any
        self._mapping[key] = value
        self._mapping[self._INDEX] = sequence, timestamps

    def _maintenance(self, sequence, timestamps):  # O(logN)
        """It will modify input sequence and timestamps in-place"""
        now = int(time.time())
        while sequence:  # Clean up expired items
            expires_at, created_at, key = sequence[0]
            if created_at <= now < expires_at:  # Then all remaining items are fresh
                break
            self._drop_indexed_entry(timestamps, sequence[0])  # It could error out
            heapq.heappop(sequence)  # Only pop it after a successful _drop_indexed_entry()
        while self._capacity is not None and len(timestamps) > self._capacity:
            self._drop_indexed_entry(timestamps, sequence[0])  # It could error out
            heapq.heappop(sequence)  # Only pop it after a successful _drop_indexed_entry()

    def _drop_indexed_entry(self, timestamps, entry):
        """For an entry came from index, drop it from timestamps and self._mapping"""
        expires_at, created_at, key = entry
        if [expires_at, created_at] == timestamps.get(key):  # So it is not an orphan
            self._mapping.pop(key, None)  # It could raise exception
            timestamps.pop(key, None)  # This would probably always succeed

    def __setitem__(self, key, value):
        """Implements the __setitem__().

        Same characteristic as :func:`~set()`,
        but use class-wide expires_in which was specified by :func:`~__init__()`.
        """
        if self._expires_in is None:
            raise ValueError("Need a numeric value for expires_in during __init__()")
        with self._lock:
            self._set(key, value, self._expires_in)

    def __getitem__(self, key):  # O(1)
        """If the item you requested already expires, KeyError will be raised."""
        self._validate_key(key)
        with self._lock:
            # Skip self._maintenance(), because it would need O(logN) time
            sequence, timestamps = self._peek()
            expires_at, created_at = timestamps[key]  # Would raise KeyError accordingly
            now = int(time.time())
            if not created_at <= now < expires_at:
                self._mapping.pop(key, None)
                timestamps.pop(key, None)
                self._mapping[self._INDEX] = sequence, timestamps
                raise KeyError("{} {}".format(
                    key,
                    "expired" if now >= expires_at else "created in the future?",
                    ))
            return self._mapping[key]  # O(1)

    def __delitem__(self, key):  # O(1)
        """If the item you requested already expires, KeyError will be raised."""
        self._validate_key(key)
        with self._lock:
            # Skip self._maintenance(), because it would need O(logN) time
            self._mapping.pop(key, None)  # O(1)
            sequence, timestamps = self._peek()
            del timestamps[key]  # O(1)
            self._mapping[self._INDEX] = sequence, timestamps

    def __len__(self):  # O(logN)
        """Drop all expired items and return the remaining length"""
        with self._lock:
            sequence, timestamps = self._peek()
            self._maintenance(sequence, timestamps)  # O(logN)
            self._mapping[self._INDEX] = sequence, timestamps
            return len(timestamps)  # Faster than iter(self._mapping) when it is on disk

    def __iter__(self):
        """Drop all expired items and return an iterator of the remaining items"""
        with self._lock:
            sequence, timestamps = self._peek()
            self._maintenance(sequence, timestamps)  # O(logN)
            self._mapping[self._INDEX] = sequence, timestamps
        return iter(timestamps)  # Faster than iter(self._mapping) when it is on disk


class _IndividualCache(object):
    # The code structure below can decorate both function and method.
    # It is inspired by https://stackoverflow.com/a/9417088
    # We may potentially switch to build upon
    # https://github.com/micheles/decorator/blob/master/docs/documentation.md#statement-of-the-problem
    def __init__(self, mapping=None, key_maker=None, expires_in=None):
        """Constructs a cache decorator that allows item-by-item control on
        how to cache the return value of the decorated function.

        :param MutableMapping mapping:
            The cached items will be stored inside.
            You'd want to use a ExpiringMapping
            if you plan to utilize the ``expires_in`` behavior.

            If nothing is provided, an in-memory dict will be used,
            but it will provide no expiry functionality.

            .. note::

                When using this class as a decorator,
                your mapping needs to be available at "compile" time,
                so it would typically be a global-, module- or class-level mapping::

                    module_mapping = {}

                    @IndividualCache(mapping=module_mapping, ...)
                    def foo():
                        ...

                If you want to use a mapping available only at run-time,
                you have to manually decorate your function at run-time, too::

                    def foo():
                        ...

                    def bar(runtime_mapping):
                        foo = IndividualCache(mapping=runtime_mapping...)(foo)

        :param callable key_maker:
            A callable which should have signature as
            ``lambda function, args, kwargs: "return a string as key"``.

            If key_maker happens to return ``None``, the cache will be bypassed,
            the underlying function will be invoked directly,
            and the invoke result will not be cached either.

        :param callable expires_in:
            The default value is ``None``,
            which means the content being cached has no per-item expiry,
            and will subject to the underlying mapping's global expiry time.

            It can be an integer indicating
            how many seconds the result will be cached.
            In particular, if the value is 0,
            it means the result expires after zero second (i.e. immediately),
            therefore the result will *not* be cached.
            (Mind the difference between ``expires_in=0`` and ``expires_in=None``.)

            Or it can be a callable with the signature as
            ``lambda function=function, args=args, kwargs=kwargs, result=result: 123``
            to calculate the expiry on the fly.
            Its return value will be interpreted in the same way as above.
        """
        self._mapping = mapping if mapping is not None else {}
        self._key_maker = key_maker or (lambda function, args, kwargs: (
            function,  # This default implementation uses function as part of key,
                # so that the cache is partitioned by function.
                # However, you could have many functions to use same namespace,
                # so different decorators could share same cache.
            args,
            tuple(kwargs.items()),  # raw kwargs is not hashable
            ))
        self._expires_in = expires_in

    def __call__(self, function):

        @wraps(function)
        def wrapper(*args, **kwargs):
            key = self._key_maker(function, args, kwargs)
            if key is None:  # Then bypass the cache
                return function(*args, **kwargs)

            now = int(time.time())
            try:
                return self._mapping[key]
            except KeyError:
                # We choose to NOT call function(...) in this block, otherwise
                # potential exception from function(...) would become a confusing
                # "During handling of the above exception, another exception occurred"
                pass
            value = function(*args, **kwargs)

            expires_in = self._expires_in(
                function=function,
                args=args,
                kwargs=kwargs,
                result=value,
                ) if callable(self._expires_in) else self._expires_in
            if expires_in == 0:
                return value
            if expires_in is None:
                self._mapping[key] = value
            else:
                self._mapping.set(key, value, expires_in)
            return value

        return wrapper




import unittest
from unittest.mock import Mock, patch
from msal.application import ConfidentialClientApplication


@patch('msal.application.Authority')
@patch('msal.application.JwtAssertionCreator', new_callable=lambda: Mock(
    return_value=Mock(create_regenerative_assertion=Mock(return_value="mock_jwt_assertion"))))
class TestClientCredentialWithOptionalThumbprint(unittest.TestCase):
    """Test that thumbprint is optional when public_certificate is provided"""

    # Sample test certificate and private key (PEM format)
    # These are minimal valid PEM structures for testing
    test_private_key = """-----BEGIN PRIVATE KEY-----
MIIEvQIBADANBgkqhkiG9w0BAQEFAASCBKcwggSjAgEAAoIBAQC7VJTUt9Us8cKj
MzEfYyjiWA4R4/M2bS1+fWIcPm15j7uo6xKvRr4PNx5bKMDFqMdW6/xfqFWX0nZK
-----END PRIVATE KEY-----"""

    test_certificate = """-----BEGIN CERTIFICATE-----
MIIC5jCCAc6gAwIBAgIJALdYQVsVsNZHMA0GCSqGSIb3DQEBCwUAMBYxFDASBgNV
BAMMC0V4YW1wbGUgQ0EwHhcNMjQwMTAxMDAwMDAwWhcNMjUwMTAxMDAwMDAwWjAW
-----END CERTIFICATE-----"""

    def _setup_mocks(self, mock_authority_class, authority="https://login.microsoftonline.com/common"):
        """Helper to setup Authority mock"""
        # Setup Authority mock
        mock_authority = Mock()
        mock_authority.is_adfs = "adfs" in authority.lower()

        # Extract instance from authority URL
        if mock_authority.is_adfs:
            # For ADFS: https://adfs.contoso.com/adfs -> adfs.contoso.com
            mock_authority.instance = authority.split("//")[1].split("/")[0]
            mock_authority.token_endpoint = f"https://{mock_authority.instance}/adfs/oauth2/token"
            mock_authority.authorization_endpoint = f"https://{mock_authority.instance}/adfs/oauth2/authorize"
        else:
            # For AAD: https://login.microsoftonline.com/common -> login.microsoftonline.com
            mock_authority.instance = authority.split("//")[1].split("/")[0]
            mock_authority.token_endpoint = f"https://{mock_authority.instance}/common/oauth2/v2.0/token"
            mock_authority.authorization_endpoint = f"https://{mock_authority.instance}/common/oauth2/v2.0/authorize"

        mock_authority.device_authorization_endpoint = None
        mock_authority_class.return_value = mock_authority

        return mock_authority

    def _setup_certificate_mocks(self, mock_extract, mock_load_cert):
        """Helper to setup certificate parsing mocks"""
        # Mock certificate loading
        mock_cert = Mock()
        mock_load_cert.return_value = mock_cert

        # Mock _extract_cert_and_thumbprints to return thumbprints
        mock_extract.return_value = (
            "mock_sha256_thumbprint",  # sha256_thumbprint
            "mock_sha1_thumbprint",     # sha1_thumbprint
            ["mock_x5c_value"]          # x5c
        )

    def _verify_assertion_params(self, mock_jwt_creator_class, expected_algorithm,
                                  expected_thumbprint_type, expected_thumbprint_value=None,
                                  has_x5c=False):
        """Helper to verify JwtAssertionCreator was called with correct params"""
        mock_jwt_creator_class.assert_called_once()
        call_args = mock_jwt_creator_class.call_args

        # Verify algorithm
        self.assertEqual(call_args[1]['algorithm'], expected_algorithm)

        # Verify thumbprint type
        if expected_thumbprint_type == 'sha256':
            self.assertIn('sha256_thumbprint', call_args[1])
            self.assertNotIn('sha1_thumbprint', call_args[1])
        elif expected_thumbprint_type == 'sha1':
            self.assertIn('sha1_thumbprint', call_args[1])
            self.assertNotIn('sha256_thumbprint', call_args[1])
            if expected_thumbprint_value:
                self.assertEqual(call_args[1]['sha1_thumbprint'], expected_thumbprint_value)

        # Verify x5c header if expected
        if has_x5c:
            self.assertIn('headers', call_args[1])
            self.assertIn('x5c', call_args[1]['headers'])

        return call_args

    @patch('cryptography.x509.load_pem_x509_certificate')
    @patch('msal.application._extract_cert_and_thumbprints')
    def test_pem_with_certificate_only_uses_sha256(
            self, mock_extract, mock_load_cert, mock_jwt_creator_class, mock_authority_class):
        """Test that providing only public_certificate (no thumbprint) uses SHA-256"""
        authority = "https://login.microsoftonline.com/common"
        self._setup_mocks(mock_authority_class, authority)
        self._setup_certificate_mocks(mock_extract, mock_load_cert)

        # Create app with certificate credential WITHOUT thumbprint
        app = ConfidentialClientApplication(
            client_id="my_client_id",
            client_credential={
                "private_key": self.test_private_key,
                "public_certificate": self.test_certificate,
                # Note: NO thumbprint provided
            },
            authority=authority
        )

        # Verify SHA-256 with PS256 algorithm is used
        self._verify_assertion_params(
            mock_jwt_creator_class,
            expected_algorithm='PS256',
            expected_thumbprint_type='sha256',
            has_x5c=True
        )

    def test_pem_with_manual_thumbprint_uses_sha1(
            self, mock_jwt_creator_class, mock_authority_class):
        """Test that providing manual thumbprint (no certificate) uses SHA-1"""
        authority = "https://login.microsoftonline.com/common"
        self._setup_mocks(mock_authority_class, authority)

        # Create app with manual thumbprint (legacy approach)
        manual_thumbprint = "A1B2C3D4E5F6"
        app = ConfidentialClientApplication(
            client_id="my_client_id",
            client_credential={
                "private_key": self.test_private_key,
                "thumbprint": manual_thumbprint,
                # Note: NO public_certificate provided
            },
            authority=authority
        )

        # Verify SHA-1 with RS256 algorithm is used
        self._verify_assertion_params(
            mock_jwt_creator_class,
            expected_algorithm='RS256',
            expected_thumbprint_type='sha1',
            expected_thumbprint_value=manual_thumbprint
        )

    def test_pem_with_both_uses_manual_thumbprint_as_sha1(
            self, mock_jwt_creator_class, mock_authority_class):
        """Test that providing both thumbprint and certificate prefers manual thumbprint (SHA-1)"""
        authority = "https://login.microsoftonline.com/common"
        self._setup_mocks(mock_authority_class, authority)

        # Create app with BOTH thumbprint and certificate
        manual_thumbprint = "A1B2C3D4E5F6"
        app = ConfidentialClientApplication(
            client_id="my_client_id",
            client_credential={
                "private_key": self.test_private_key,
                "thumbprint": manual_thumbprint,
                "public_certificate": self.test_certificate,
            },
            authority=authority
        )

        # Verify manual thumbprint takes precedence (backward compatibility)
        self._verify_assertion_params(
            mock_jwt_creator_class,
            expected_algorithm='RS256',
            expected_thumbprint_type='sha1',
            expected_thumbprint_value=manual_thumbprint,
            has_x5c=True  # x5c should still be present
        )

    @patch('cryptography.x509.load_pem_x509_certificate')
    @patch('msal.application._extract_cert_and_thumbprints')
    def test_pem_with_adfs_uses_sha1(
            self, mock_extract, mock_load_cert, mock_jwt_creator_class, mock_authority_class):
        """Test that ADFS authority uses SHA-1 even with SHA-256 thumbprint"""
        authority = "https://adfs.contoso.com/adfs"
        self._setup_mocks(mock_authority_class, authority)
        self._setup_certificate_mocks(mock_extract, mock_load_cert)

        # Create app with certificate on ADFS
        app = ConfidentialClientApplication(
            client_id="my_client_id",
            client_credential={
                "private_key": self.test_private_key,
                "public_certificate": self.test_certificate,
            },
            authority=authority
        )

        # ADFS should force SHA-1 with RS256 even though SHA-256 would be calculated
        self._verify_assertion_params(
            mock_jwt_creator_class,
            expected_algorithm='RS256',
            expected_thumbprint_type='sha1'
        )

    def test_pem_with_neither_raises_error(self, mock_jwt_creator_class, mock_authority_class):
        """Test that providing neither thumbprint nor certificate raises ValueError"""
        authority = "https://login.microsoftonline.com/common"
        self._setup_mocks(mock_authority_class, authority)

        # Should raise ValueError when neither thumbprint nor certificate provided
        with self.assertRaises(ValueError) as context:
            app = ConfidentialClientApplication(
                client_id="my_client_id",
                client_credential={
                    "private_key": self.test_private_key,
                    # Note: NO thumbprint and NO public_certificate
                },
                authority=authority
            )

        self.assertIn("thumbprint", str(context.exception).lower())
        self.assertIn("public_certificate", str(context.exception).lower())


if __name__ == "__main__":
    unittest.main()



#------------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation.
# All rights reserved.
#
# This code is licensed under the MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions :
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#------------------------------------------------------------------------------

try:
    from xml.etree import cElementTree as ET
except ImportError:
    from xml.etree import ElementTree as ET
import os

from msal.wstrust_request import _build_rst, escape_xml
from msal.wstrust_response import *

from tests import unittest


class Test_WsTrustResponse(unittest.TestCase):

    def test_findall_content_with_comparison(self):
        content = """
            <saml:Assertion xmlns:saml="SAML:assertion">
                <ds:Signature xmlns:ds="http://www.w3.org/2000/09/xmldsig#">
                foo
                </ds:Signature>
            </saml:Assertion>"""
        sample = ('<ns0:Wrapper xmlns:ns0="namespace0">'
            + content
            + '</ns0:Wrapper>')

        # Demonstrating how XML-based parser won't give you the raw content as-is
        element = ET.fromstring(sample).findall('{SAML:assertion}Assertion')[0]
        assertion_via_xml_parser = ET.tostring(element)
        self.assertNotEqual(content, assertion_via_xml_parser)
        self.assertNotIn(b"<ds:Signature>", assertion_via_xml_parser)

        # The findall_content() helper, based on Regex, will return content as-is.
        self.assertEqual([content], findall_content(sample, "Wrapper"))

    def test_parse_error(self):
        error_response = '''
            <s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope" xmlns:a="http://www.w3.org/2005/08/addressing" xmlns:u="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd">
              <s:Header>
               <a:Action s:mustUnderstand="1">http://www.w3.org/2005/08/addressing/soap/fault</a:Action>
               <o:Security s:mustUnderstand="1" xmlns:o="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd">
                 <u:Timestamp u:Id="_0">
                 <u:Created>2013-07-30T00:32:21.989Z</u:Created>
                 <u:Expires>2013-07-30T00:37:21.989Z</u:Expires>
                 </u:Timestamp>
               </o:Security>
               </s:Header>
             <s:Body>
               <s:Fault>
                 <s:Code>
                   <s:Value>s:Sender</s:Value>
                   <s:Subcode>
                   <s:Value xmlns:a="http://docs.oasis-open.org/ws-sx/ws-trust/200512">a:RequestFailed</s:Value>
                   </s:Subcode>
                 </s:Code>
                 <s:Reason>
                 <s:Text xml:lang="en-US">MSIS3127: The specified request failed.</s:Text>
                 </s:Reason>
               </s:Fault>
            </s:Body>
            </s:Envelope>'''
        self.assertEqual({
            "reason": "MSIS3127: The specified request failed.",
            "code": "a:RequestFailed",
            }, parse_error(error_response))

    def test_token_parsing_happy_path(self):
        with open(os.path.join(os.path.dirname(__file__), "rst_response.xml")) as f:
            rst_body = f.read()
        result = parse_token_by_re(rst_body)
        self.assertEqual(result.get("type"), SAML_TOKEN_TYPE_V1)
        self.assertIn(b"<saml:Assertion", result.get("token", ""))


class Test_WsTrustRequest(unittest.TestCase):

    def test_escape_xml(self):
        self.assertEqual(escape_xml('<>&"\''), '&lt;&gt;&amp;&quot;&apos;')

    def test_username_xml_injection_is_prevented(self):
        malicious = 'admin</wsse:Username><x>INJECTED'
        rst = _build_rst(malicious, 'pw', 'urn:x', 'https://x',
            'http://docs.oasis-open.org/ws-sx/ws-trust/200512/RST/Issue')
        self.assertEqual(rst.count('<wsse:Username>'), 1)
        self.assertNotIn('<x>INJECTED', rst)




from time import sleep
from random import random
import unittest
from msal.individual_cache import _ExpiringMapping as ExpiringMapping
from msal.individual_cache import _IndividualCache as IndividualCache


class TestExpiringMapping(unittest.TestCase):
    def setUp(self):
        self.mapping = {}
        self.expires_in = 1
        self.m = ExpiringMapping(
            mapping=self.mapping, capacity=2, expires_in=self.expires_in)

    def how_many(self):
        # This helper checks how many items are in the mapping, WITHOUT triggering purge
        return len(self.m._peek()[1])

    def test_should_disallow_accessing_reserved_keyword(self):
        with self.assertRaises(ValueError):
            self.m.get(ExpiringMapping._INDEX)

    def test_setitem(self):
        self.assertEqual(0, len(self.m))
        self.m["thing one"] = "one"
        self.assertIn(ExpiringMapping._INDEX, self.mapping, "Index created")
        self.assertEqual(1, len(self.m), "It contains one item (excluding index)")
        self.assertEqual("one", self.m["thing one"])
        self.assertEqual(["thing one"], list(self.m))

    def test_set(self):
        self.assertEqual(0, len(self.m))
        self.m.set("thing two", "two", 2)
        self.assertIn(ExpiringMapping._INDEX, self.mapping, "Index created")
        self.assertEqual(1, len(self.m), "It contains one item (excluding index)")
        self.assertEqual("two", self.m["thing two"])
        self.assertEqual(["thing two"], list(self.m))

    def test_len_should_purge(self):
        self.m["thing one"] = "one"
        sleep(1)
        self.assertEqual(0, len(self.m))

    def test_iter_should_purge(self):
        self.m["thing one"] = "one"
        sleep(1)
        self.assertEqual([], list(self.m))

    def test_get_should_not_purge_and_should_return_only_when_the_item_is_still_valid(self):
        self.m["thing one"] = "one"
        self.m["thing two"] = "two"
        sleep(1)
        self.assertEqual(2, self.how_many(), "We begin with 2 items")
        with self.assertRaises(KeyError):
            self.m["thing one"]
        self.assertEqual(1, self.how_many(), "get() should not purge the remaining items")

    def test_setitem_should_purge(self):
        self.m["thing one"] = "one"
        sleep(1)
        self.m["thing two"] = "two"
        self.assertEqual(1, self.how_many(), "setitem() should purge all expired items")
        self.assertEqual("two", self.m["thing two"], "The remaining item should be thing two")

    def test_various_expiring_time(self):
        self.assertEqual(0, len(self.m))
        self.m["thing one"] = "one"
        self.m.set("thing two", "two", 2)
        self.assertEqual(2, len(self.m), "It contains 2 items")
        sleep(1)
        self.assertEqual(["thing two"], list(self.m), "One expires, another remains")

    def test_old_item_can_be_updated_with_new_expiry_time(self):
        self.assertEqual(0, len(self.m))
        self.m["thing"] = "one"
        new_lifetime = 3  # 2-second seems too short and causes flakiness
        self.m.set("thing", "two", new_lifetime)
        self.assertEqual(1, len(self.m), "It contains 1 item")
        self.assertEqual("two", self.m["thing"], 'Already been updated to "two"')
        sleep(self.expires_in)
        self.assertEqual("two", self.m["thing"], "Not yet expires")
        sleep(new_lifetime - self.expires_in)
        self.assertEqual(0, len(self.m))

    def test_oversized_input_should_purge_most_aging_item(self):
        self.assertEqual(0, len(self.m))
        self.m["thing one"] = "one"
        self.m.set("thing two", "two", 2)
        self.assertEqual(2, len(self.m), "It contains 2 items")
        self.m["thing three"] = "three"
        self.assertEqual(2, len(self.m), "It contains 2 items")
        self.assertNotIn("thing one", self.m)


class TestIndividualCache(unittest.TestCase):
    mapping = {}

    @IndividualCache(mapping=mapping)
    def foo(self, a, b, c=None, d=None):
        return random()  # So that we'd know whether a new response is received

    def test_memorize_a_function_call(self):
        self.assertNotEqual(self.foo(1, 1), self.foo(2, 2))
        self.assertEqual(
            self.foo(1, 2, c=3, d=4),
            self.foo(1, 2, c=3, d=4),
            "Subsequent run should obtain same result from cache")
        # Note: In Python 3.7+, dict is ordered, so the following is typically True:
        #self.assertNotEqual(self.foo(a=1, b=2), self.foo(b=2, a=1))




import json

from msal.oauth2cli import JwtAssertionCreator
from msal.oauth2cli.oidc import decode_part

from tests import unittest


class AssertionTestCase(unittest.TestCase):
    def test_extra_claims(self):
        assertion = JwtAssertionCreator(key=None, algorithm="none").sign_assertion(
            "audience", "issuer", additional_claims={"client_ip": "1.2.3.4"})
        payload = json.loads(decode_part(assertion.split(b'.')[1].decode('utf-8')))
        self.assertEqual("1.2.3.4", payload.get("client_ip"))

